"""
Module for communication with Sentinel Hub Opensearch service.

For more search parameters check
`service description <http://opensearch.sentinel-hub.com/resto/api/collections/Sentinel2/describe.xml>`__.
"""
import datetime as dt
import logging
from typing import Iterable, Iterator, List, Optional, Union
from urllib.parse import urlencode

from ..config import SHConfig
from ..constants import CRS
from ..download import DownloadClient
from ..geometry import BBox
from ..time_utils import RawTimeIntervalType, RawTimeType, parse_time, parse_time_interval, serialize_time
from ..type_utils import JsonDict

LOGGER = logging.getLogger(__name__)


class TileMissingException(Exception):
    """This exception is raised when requested tile is missing at Sentinel Hub Opensearch service"""


def get_tile_info_id(tile_id: str) -> JsonDict:
    """Get basic information about image tile

    :param tile_id: original tile identification string provided by ESA (e.g.
        'S2A_OPER_MSI_L1C_TL_SGS__20160109T230542_A002870_T10UEV_N02.01')
    :return: dictionary with info provided by Opensearch REST service
    :raises: TileMissingException if no tile with tile ID `tile_id` exists
    """
    result_list = list(search_iter(tile_id=tile_id))

    if not result_list:
        raise TileMissingException

    if len(result_list) > 1:
        LOGGER.warning("Obtained %d results for tile_id=%s. Returning the first one", len(result_list), tile_id)

    return result_list[0]


def get_tile_info(
    tile: str, time: Union[RawTimeType, RawTimeIntervalType], aws_index: Optional[int] = None, all_tiles: bool = False
) -> Union[JsonDict, List[JsonDict]]:
    """Get basic information about image tile

    :param tile: tile name (e.g. ``'T10UEV'``)
    :param time: A single date or a time interval
    :param aws_index: index of tile on AWS
    :param all_tiles: If `True` it will return list of all tiles otherwise only the first one
    :return: dictionary (or list of dictionaries) with info provided by Opensearch REST service
    """
    start_date, end_date = parse_time_interval(time)

    candidates = []
    for tile_info in search_iter(start_date=start_date, end_date=end_date):
        path_props = tile_info["properties"]["s3Path"].split("/")
        this_tile = "".join(path_props[1:4])
        this_aws_index = int(path_props[-1])
        if this_tile == tile.lstrip("T0") and (aws_index is None or aws_index == this_aws_index):
            candidates.append(tile_info)

        if candidates and aws_index is not None:
            break
        if len(candidates) >= 2 and not all_tiles:
            break

    if not candidates:
        raise TileMissingException
    if all_tiles:
        return candidates

    if len(candidates) > 1:
        LOGGER.info("Obtained more than one result for tile=%s, time=%s. Returning the first one", tile, time)
    return candidates[0]


def get_area_info(bbox: BBox, date_interval: RawTimeIntervalType, maxcc: Optional[float] = None) -> List[JsonDict]:
    """Get information about all images from specified area and time range

    :param bbox: bounding box of requested area
    :param date_interval: a pair of time strings in ISO8601 format
    :param maxcc: filter images by maximum percentage of cloud coverage
    :return: iterator of dictionaries containing info provided by Opensearch REST service
    """
    result_list = search_iter(bbox=bbox, start_date=date_interval[0], end_date=date_interval[1])
    if maxcc:
        return reduce_by_maxcc(result_list, maxcc)
    return list(result_list)


def get_area_dates(bbox: BBox, date_interval: RawTimeIntervalType, maxcc: Optional[int] = None) -> List[dt.date]:
    """Get list of times of existing images from specified area and time range

    :param bbox: bounding box of requested area
    :param date_interval: a pair of time strings in ISO8601 format
    :param maxcc: filter images by maximum percentage of cloud coverage
    :return: list of time strings in ISO8601 format
    """

    area_info = get_area_info(bbox, date_interval, maxcc=maxcc)
    return sorted({parse_time(tile_info["properties"]["startDate"]) for tile_info in area_info})  # type: ignore


def reduce_by_maxcc(result_list: Iterable[JsonDict], maxcc: float) -> List[JsonDict]:
    """Filter list image tiles by maximum cloud coverage

    :param result_list: list of dictionaries containing info provided by Opensearch REST service
    :param maxcc: filter images by maximum percentage of cloud coverage
    :return: list of dictionaries containing info provided by Opensearch REST service
    """
    return [tile_info for tile_info in result_list if tile_info["properties"]["cloudCover"] <= 100 * float(maxcc)]


def search_iter(
    tile_id: Optional[str] = None,
    bbox: Optional[BBox] = None,
    start_date: Optional[RawTimeType] = None,
    end_date: Optional[RawTimeType] = None,
    absolute_orbit: Optional[int] = None,
    config: Optional[SHConfig] = None,
) -> Iterator[JsonDict]:
    """A generator function that implements OpenSearch search queries and returns results

    All parameters for search are optional.

    :param tile_id: original tile identification string provided by ESA (e.g.
                    'S2A_OPER_MSI_L1C_TL_SGS__20160109T230542_A002870_T10UEV_N02.01')
    :param bbox: bounding box of requested area
    :param start_date: beginning of time range
    :param end_date: end of time range
    :param absolute_orbit: An absolute orbit number of Sentinel-2 L1C products as defined by ESA
    :return: An iterator returning dictionaries with info provided by Sentinel Hub OpenSearch REST service
    :param config: A custom instance of config class to override parameters from the saved configuration.
    """
    config = config or SHConfig()

    if bbox and bbox.crs is not CRS.WGS84:
        bbox = bbox.transform(CRS.WGS84)

    start_date = parse_time(start_date) if start_date else None
    end_date = parse_time(end_date) if end_date else None

    url_params = _prepare_url_params(tile_id, bbox, end_date, start_date, absolute_orbit)
    url_params["maxRecords"] = config.max_opensearch_records_per_query

    start_index = 1
    client = DownloadClient(config=config)

    while True:
        url_params["index"] = start_index

        url = f"{config.opensearch_url}/search.json?{urlencode(url_params)}"
        LOGGER.debug("URL=%s", url)

        response = client.get_json_dict(url)
        for tile_info in response["features"]:
            yield tile_info

        if len(response["features"]) < config.max_opensearch_records_per_query:
            break
        start_index += config.max_opensearch_records_per_query


def _prepare_url_params(
    tile_id: Optional[str],
    bbox: Optional[BBox],
    end_date: Optional[dt.date],
    start_date: Optional[dt.date],
    absolute_orbit: Optional[int],
) -> JsonDict:
    """Constructs dict with URL params

    :param tile_id: original tile identification string provided by ESA (e.g.
                    'S2A_OPER_MSI_L1C_TL_SGS__20160109T230542_A002870_T10UEV_N02.01')
    :param bbox: bounding box of requested area in WGS84 CRS
    :param start_date: beginning of time range
    :param end_date: end of time range
    :param absolute_orbit: An absolute orbit number of Sentinel-2 L1C products as defined by ESA
    :return: dictionary with parameters as properties when arguments not None
    """
    url_params = {
        "identifier": tile_id,
        "startDate": serialize_time(start_date, use_tz=False) if start_date else None,
        "completionDate": serialize_time(end_date, use_tz=False) if end_date else None,
        "orbitNumber": absolute_orbit,
        "box": bbox,
    }
    return {key: str(value) for key, value in url_params.items() if value}



"""
Implementation of
`Sentinel Hub Statistical API interface <https://docs.sentinel-hub.com/api/latest/api/statistical/>`__.
"""
from typing import Any, List, Optional, Tuple, Union

from ..constants import MimeType
from ..download.sentinelhub_statistical_client import SentinelHubStatisticalDownloadClient
from ..geometry import BBox, Geometry
from ..time_utils import parse_time_interval, serialize_time
from ..type_utils import JsonDict, RawTimeIntervalType
from .base_request import InputDataDict, SentinelHubBaseApiRequest
from .utils import _update_other_args


class SentinelHubStatistical(SentinelHubBaseApiRequest):
    """Sentinel Hub Statistical API interface

    For more information check
    `Statistical API documentation <https://docs.sentinel-hub.com/api/latest/api/statistical/>`__.
    """

    _SERVICE_ENDPOINT = "statistics"

    def __init__(
        self,
        aggregation: JsonDict,
        input_data: List[Union[JsonDict, InputDataDict]],
        bbox: Optional[BBox] = None,
        geometry: Optional[Geometry] = None,
        calculations: Optional[JsonDict] = None,
        **kwargs: Any
    ):
        """
        For details of certain parameters check the
        `Statistical API reference <https://docs.sentinel-hub.com/api/latest/reference/#tag/statistical>`_.

        :param aggregation: Aggregation part of the payload, which can be generated with `aggregation` method
        :param input_data: A list of input dictionary objects as described in the API reference. It can be generated
            with `input_data` method
        :param bbox: A bounding box of the request
        :param geometry: A geometry of the request
        :param calculations: Calculations part of the payload.
        :param data_folder: Location of the directory where the downloaded data could be saved.
        :param config: A custom instance of config class to override parameters from the saved configuration.
        """
        self.payload = self.body(
            request_bounds=self.bounds(bbox=bbox, geometry=geometry),
            request_data=input_data,
            aggregation=aggregation,
            calculations=calculations,
        )

        super().__init__(SentinelHubStatisticalDownloadClient, **kwargs)

    @property
    def mime_type(self) -> MimeType:
        return MimeType.JSON

    @staticmethod
    def body(
        request_bounds: JsonDict,
        request_data: List[JsonDict],
        aggregation: JsonDict,
        calculations: Optional[JsonDict],
        other_args: Optional[JsonDict] = None,
    ) -> JsonDict:
        """Generate the Process API request body

        :param request_bounds: A dictionary as generated by `bounds` helper method.
        :param request_data: A list of dictionaries as generated by `input_data` helper method.
        :param aggregation: A dictionary as generated by `aggregation` helper method.
        :param calculations: A dictionary defining calculations part of the payload
        :param other_args: Additional dictionary of arguments. If provided, the resulting dictionary will get updated
            by it.
        :returns: Request payload dictionary
        """
        # Some parts of the payload have to be defined:
        for input_data_payload in request_data:
            if "dataFilter" not in input_data_payload:
                input_data_payload["dataFilter"] = {}

        if calculations is None:
            calculations = {"default": {}}

        request_body = {
            "input": {"bounds": request_bounds, "data": request_data},
            "aggregation": aggregation,
            "calculations": calculations,
        }

        if other_args:
            _update_other_args(request_body, other_args)

        return request_body

    @staticmethod
    def aggregation(
        evalscript: str,
        time_interval: RawTimeIntervalType,
        aggregation_interval: str,
        size: Optional[Tuple[int, int]] = None,
        resolution: Optional[Tuple[float, float]] = None,
        other_args: Optional[JsonDict] = None,
    ) -> JsonDict:
        """Generate the `aggregation` part of the Statistical API request body

        :param evalscript: An `evalscript <https://docs.sentinel-hub.com/api/latest/#/Evalscript/>`_.
        :param time_interval: An interval with start and end date of the form YYYY-MM-DDThh:mm:ss or YYYY-MM-DD or
            a datetime object
        :param aggregation_interval: How data from given time interval is aggregated together
        :param size: A width and height of an image from which data will be aggregated
        :param resolution: A resolution in x and y dimensions of an image from which data will be aggregated.
            Resolution has to be defined in the same units as request bbox or geometry
        :param other_args: Additional dictionary of arguments. If provided, the resulting dictionary will get updated
            by it.
        :returns: Aggregation payload dictionary
        """
        start_time, end_time = serialize_time(parse_time_interval(time_interval, allow_undefined=True), use_tz=True)

        payload: JsonDict = {
            "evalscript": evalscript,
            "timeRange": {"from": start_time, "to": end_time},
            "aggregationInterval": {"of": aggregation_interval},
        }

        if size:
            payload["width"], payload["height"] = size
        if resolution:
            payload["resx"], payload["resy"] = resolution

        if other_args:
            _update_other_args(payload, other_args)

        return payload



"""
Data request interface for Geopedia services
"""

from abc import abstractmethod
from typing import Any, Optional

from ..base import DataRequest
from ..constants import CRS, MimeType, ServiceType
from ..download import DownloadClient
from ..geometry import BBox
from .core import GeopediaFeatureIterator, GeopediaImageService, GeopediaSession, GeopediaWmsService


class GeopediaRequest(DataRequest):
    """The base class for Geopedia requests where all common parameters are defined."""

    def __init__(
        self,
        layer: str,
        service_type: ServiceType,
        *,
        bbox: BBox,
        theme: Optional[str] = None,
        image_format: MimeType = MimeType.PNG,
        **kwargs: Any,
    ):
        """
        :param layer: Geopedia layer which contains requested data
        :param service_type: Type of the service, supported are ``ServiceType.WMS`` and ``ServiceType.IMAGE``
        :param bbox: Bounding box of the requested data
        :param theme: Geopedia's theme endpoint string for which the layer is defined. Only required by WMS service.
        :param image_format: Format of the returned image by the Sentinel Hub's WMS getMap service. Default is
            ``constants.MimeType.PNG``.
        :param data_folder: Location of the directory where the fetched data will be saved.
        :param config: A custom instance of config class to override parameters from the saved configuration.
        """
        self.layer = layer
        self.service_type = service_type

        self.bbox = bbox
        if bbox.crs is not CRS.POP_WEB:
            raise ValueError(
                f"Geopedia Request at the moment supports only bounding boxes with coordinates in {CRS.POP_WEB}"
            )

        self.theme = theme
        self.image_format = MimeType(image_format)

        super().__init__(DownloadClient, **kwargs)

    @abstractmethod
    def create_request(self) -> None:
        raise NotImplementedError


class GeopediaWmsRequest(GeopediaRequest):
    """Web Map Service request class for Geopedia

    Creates an instance of Geopedia's WMS (Web Map Service) GetMap request, which provides access to WMS layers in
    Geopedia.
    """

    def __init__(
        self,
        layer: str,
        theme: str,
        bbox: BBox,
        *,
        width: Optional[int] = None,
        height: Optional[int] = None,
        **kwargs: Any,
    ):
        """
        :param layer: Geopedia layer which contains requested data
        :param theme: Geopedia's theme endpoint string for which the layer is defined.
        :param bbox: Bounding box of the requested data
        :param width: width (number of columns) of the returned image (array)
        :param height: height (number of rows) of the returned image (array)
        :param image_format: Format of the returned image by the Sentinel Hub's WMS getMap service. Default is
            ``constants.MimeType.PNG``.
        :param data_folder: Location of the directory where the fetched data will be saved.
        :param config: A custom instance of config class to override parameters from the saved configuration.
        """
        self.size_x = width
        self.size_y = height

        super().__init__(layer=layer, theme=theme, bbox=bbox, service_type=ServiceType.WMS, **kwargs)

    def create_request(self) -> None:
        """Set download requests

        Create a list of DownloadRequests for all Sentinel-2 acquisitions within request's time interval and
        acceptable cloud coverage.
        """
        gpd_service = GeopediaWmsService(config=self.config)
        self.download_list = gpd_service.get_request(self)  # type: ignore[arg-type]


class GeopediaImageRequest(GeopediaRequest):
    """Request to access data in a Geopedia vector / raster layer."""

    def __init__(
        self,
        *,
        image_field_name: str,
        keep_image_names: bool = True,
        gpd_session: Optional[GeopediaSession] = None,
        **kwargs: Any,
    ):
        """
        :param image_field_name: Name of the field in the data table which holds images
        :param keep_image_names: If `True` images will be saved with the same names as in Geopedia otherwise Geopedia
            hashes will be used as names. If there are multiple images with the same names in the Geopedia layer this
            parameter should be set to `False` to prevent images being overwritten.
        :param layer: Geopedia layer which contains requested data
        :param bbox: Bounding box of the requested data
        :param image_format: Format of the returned image by the Sentinel Hub's WMS getMap service. Default is
            ``constants.MimeType.PNG``.
        :param gpd_session: Optional parameter for specifying a custom Geopedia session, which can also contain login
            credentials. This can be used for accessing private Geopedia layers. By default, it is set to `None` and a
            basic Geopedia session without credentials will be created.
        :param data_folder: Location of the directory where the fetched data will be saved.
        :param config: A custom instance of config class to override parameters from the saved configuration.
        """
        self.image_field_name = image_field_name
        self.keep_image_names = keep_image_names
        self.gpd_session = gpd_session

        self.gpd_iterator: Optional[GeopediaFeatureIterator] = None

        super().__init__(service_type=ServiceType.IMAGE, **kwargs)

    def create_request(self, reset_gpd_iterator: bool = False) -> None:
        """Set a list of download requests

        Set a list of DownloadRequests for all images that are under the
        given property of the Geopedia's Vector layer.

        :param reset_gpd_iterator: When re-running the method this flag is used to reset/keep existing ``gpd_iterator``
            (i.e. instance of ``GeopediaFeatureIterator`` class). If the iterator is not reset you don't have to
            repeat a service call but tiles and dates will stay the same.
        """
        if reset_gpd_iterator:
            self.gpd_iterator = None

        gpd_service = GeopediaImageService(config=self.config)
        self.download_list = gpd_service.get_request(self)
        self.gpd_iterator = gpd_service.get_gpd_iterator()

    def get_items(self) -> Optional[GeopediaFeatureIterator]:
        """Returns iterator over info about data used for this request

        :return: Iterator of dictionaries containing info about data used in
                 this request.
        """
        return self.gpd_iterator



"""
Module implementing an interface with
`Sentinel Hub Batch Processing API <https://docs.sentinel-hub.com/api/latest/api/batch/>`__.
"""
import datetime as dt
import logging
import time
from collections import defaultdict
from dataclasses import dataclass, field
from enum import Enum
from typing import Any, DefaultDict, Dict, Iterable, Iterator, List, Optional, Tuple, Union

from dataclasses_json import CatchAll, LetterCase, Undefined
from dataclasses_json import config as dataclass_config
from dataclasses_json import dataclass_json
from tqdm.auto import tqdm

from ..config import SHConfig
from ..constants import RequestType
from ..data_collections import DataCollection
from ..geometry import CRS, BBox, Geometry
from ..type_utils import Json, JsonDict
from .base import BaseCollection, SentinelHubFeatureIterator, SentinelHubService
from .process import SentinelHubRequest
from .utils import datetime_config, enum_config, remove_undefined

LOGGER = logging.getLogger(__name__)

BatchRequestType = Union[str, dict, "BatchRequest"]
BatchCollectionType = Union[str, dict, "BatchCollection"]


class SentinelHubBatch(SentinelHubService):
    """An interface class for Sentinel Hub Batch API

    For more info check `Batch API reference <https://docs.sentinel-hub.com/api/latest/reference/#tag/batch_process>`__.
    """

    # pylint: disable=too-many-public-methods
    @staticmethod
    def _get_service_url(base_url: str) -> str:
        """Provides URL to Catalog API"""
        return f"{base_url}/api/v1/batch"

    def create(
        self,
        sentinelhub_request: Union[SentinelHubRequest, Dict[str, Any]],
        tiling_grid: Dict[str, Any],
        output: Optional[Dict[str, Any]] = None,
        bucket_name: Optional[str] = None,
        description: Optional[str] = None,
        **kwargs: Any,
    ) -> "BatchRequest":
        """Create a new batch request

        `Batch API reference
        <https://docs.sentinel-hub.com/api/latest/reference/#operation/createNewBatchProcessingRequest>`__

        :param sentinelhub_request: An instance of SentinelHubRequest class containing all request parameters.
            Alternatively, it can also be just a payload dictionary for Process API request
        :param tiling_grid: A dictionary with tiling grid parameters. It can be built with `tiling_grid` method
        :param output: A dictionary with output parameters. It can be built with `output` method. Alternatively, one
            can set `bucket_name` parameter instead.
        :param bucket_name: A name of an S3 bucket where to save data. Alternatively, one can set `output` parameter
            to specify more output parameters.
        :param description: A description of a batch request
        :param kwargs: Any other arguments to be added to a dictionary of parameters.
        :return: An instance of `SentinelHubBatch` object that represents a newly created batch request.
        """
        if isinstance(sentinelhub_request, SentinelHubRequest):
            sentinelhub_request = sentinelhub_request.download_list[0].post_values  # type: ignore[assignment]

        if not isinstance(sentinelhub_request, dict):
            raise ValueError(
                "Parameter sentinelhub_request should be an instance of SentinelHubRequest or a "
                "dictionary with a request payload"
            )

        payload = {
            "processRequest": sentinelhub_request,
            "tilingGrid": tiling_grid,
            "output": output,
            "bucketName": bucket_name,
            "description": description,
            **kwargs,
        }
        payload = remove_undefined(payload)

        url = self._get_process_url()
        request_info = self.client.get_json(url, post_values=payload, use_session=True)

        return BatchRequest.from_dict(request_info)  # type: ignore[attr-defined]

    @staticmethod
    def tiling_grid(
        grid_id: int, resolution: float, buffer: Optional[Tuple[int, int]] = None, **kwargs: Any
    ) -> JsonDict:
        """A helper method to build a dictionary with tiling grid parameters

        :param grid_id: An ID of a tiling grid
        :param resolution: A grid resolution
        :param buffer: Optionally, a buffer around each tile can be defined. It can be defined with a tuple of integers
            `(buffer_x, buffer_y)`, which specifies a number of buffer pixels in horizontal and vertical directions.
        :param kwargs: Any other arguments to be added to a dictionary of parameters
        :return: A dictionary with parameters
        """
        payload = {"id": grid_id, "resolution": resolution, **kwargs}
        if buffer:
            payload = {**payload, "bufferX": buffer[0], "bufferY": buffer[1]}
        return payload

    @staticmethod
    def output(
        *,
        default_tile_path: Optional[str] = None,
        overwrite: Optional[bool] = None,
        skip_existing: Optional[bool] = None,
        cog_output: Optional[bool] = None,
        cog_parameters: Optional[Dict[str, Any]] = None,
        create_collection: Optional[bool] = None,
        collection_id: Optional[str] = None,
        responses: Optional[List[str]] = None,
        **kwargs: Any,
    ) -> Dict[str, Any]:
        """A helper method to build a dictionary with tiling grid parameters

        :param default_tile_path: A path or a template on an s3 bucket where to store results. More info at Batch API
            documentation
        :param overwrite: A flag specifying if a request should overwrite existing outputs without failing
        :param skip_existing: A flag specifying if existing outputs should be overwritten
        :param cog_output: A flag specifying if outputs should be written in COGs (cloud-optimized GeoTIFFs )or
            normal GeoTIFFs
        :param cog_parameters: A dictionary specifying COG creation parameters
        :param create_collection: If True the results will be written in COGs and a batch collection will be created
        :param collection_id: If True results will be added to an existing collection
        :param responses: Specification of path template for individual outputs/responses
        :param kwargs: Any other arguments to be added to a dictionary of parameters
        :return: A dictionary of output parameters
        """
        return remove_undefined(
            {
                "defaultTilePath": default_tile_path,
                "overwrite": overwrite,
                "skipExisting": skip_existing,
                "cogOutput": cog_output,
                "cogParameters": cog_parameters,
                "createCollection": create_collection,
                "collectionId": collection_id,
                "responses": responses,
                **kwargs,
            }
        )

    def iter_tiling_grids(self, **kwargs: Any) -> SentinelHubFeatureIterator:
        """An iterator over tiling grids

        `Batch API reference
        <https://docs.sentinel-hub.com/api/latest/reference/#operation/getBatchTilingGridsProperties>`__

        :param kwargs: Any other request query parameters
        :return: An iterator over tiling grid definitions
        """
        return SentinelHubFeatureIterator(
            client=self.client,
            url=self._get_tiling_grids_url(),
            params=remove_undefined(kwargs),
            exception_message="Failed to obtain information about available tiling grids",
        )

    def get_tiling_grid(self, grid_id: Union[int, str]) -> JsonDict:
        """Provides a single tiling grid

        `Batch API reference
        <https://docs.sentinel-hub.com/api/latest/reference/#operation/getBatchTilingGridProperties>`__

        :param grid_id: An ID of a requested tiling grid
        :return: A tiling grid definition
        """
        url = self._get_tiling_grids_url(grid_id)
        return self.client.get_json_dict(url=url, use_session=True)

    def iter_requests(
        self, user_id: Optional[str] = None, search: Optional[str] = None, sort: Optional[str] = None, **kwargs: Any
    ) -> Iterator["BatchRequest"]:
        """Iterate existing batch requests

        `Batch API reference
        <https://docs.sentinel-hub.com/api/latest/reference/#operation/getAllBatchProcessRequests>`__

        :param user_id: Filter requests by a user id who defined a request
        :param search: A search query to filter requests
        :param sort: A sort query
        :param kwargs: Any additional parameters to include in a request query
        :return: An iterator over existing batch requests
        """
        params = remove_undefined({"userid": user_id, "search": search, "sort": sort, **kwargs})
        feature_iterator = SentinelHubFeatureIterator(
            client=self.client, url=self._get_process_url(), params=params, exception_message="No requests found"
        )
        for request_info in feature_iterator:
            yield BatchRequest.from_dict(request_info)  # type: ignore[attr-defined]

    def get_latest_request(self) -> "BatchRequest":
        """Provides a batch request that has been created the latest

        :return: Batch request info
        """
        latest_request_iter = self.iter_requests(sort="created:desc", count=1)
        try:
            return next(latest_request_iter)
        except StopIteration as exception:
            raise ValueError("No batch request is available") from exception

    def get_request(self, batch_request: BatchRequestType) -> "BatchRequest":
        """Collects information about a single batch request

        `Batch API reference
        <https://docs.sentinel-hub.com/api/latest/reference/#operation/getSingleBatchProcessRequestById>`__

        :return: Batch request info
        """
        request_id = self._parse_request_id(batch_request)
        request_info = self.client.get_json(url=self._get_process_url(request_id), use_session=True)
        return BatchRequest.from_dict(request_info)  # type: ignore[attr-defined]

    def update_request(
        self,
        batch_request: BatchRequestType,
        output: Optional[Dict[str, Any]] = None,
        description: Optional[str] = None,
        **kwargs: Any,
    ) -> Json:
        """Update batch job request parameters

        `Batch API reference
        <https://docs.sentinel-hub.com/api/latest/reference/#operation/updateBatchProcessRequest>`__

        Similarly to `update_info` method, this method also updates local information in the current instance of
        `SentinelHubBatch`.

        :param batch_request: It could be a batch request object, a raw batch request payload or only a batch
            request ID.
        :param output: A dictionary with output parameters to be updated.
        :param description: A description of a batch request to be updated.
        :param kwargs: Any other arguments to be added to a dictionary of parameters.
        """
        request_id = self._parse_request_id(batch_request)
        payload = remove_undefined({"output": output, "description": description, **kwargs})
        return self.client.get_json(
            url=self._get_process_url(request_id), post_values=payload, request_type=RequestType.PUT, use_session=True
        )

    def delete_request(self, batch_request: BatchRequestType) -> Json:
        """Delete a batch job request

        `Batch API reference
        <https://docs.sentinel-hub.com/api/latest/reference/#operation/deleteBatchProcessRequest>`__

        :param batch_request: It could be a batch request object, a raw batch request payload or only a batch
            request ID.
        """
        request_id = self._parse_request_id(batch_request)
        return self.client.get_json(
            url=self._get_process_url(request_id), request_type=RequestType.DELETE, use_session=True
        )

    def start_analysis(self, batch_request: BatchRequestType) -> Json:
        """Starts analysis of a batch job request

        `Batch API reference <https://docs.sentinel-hub.com/api/latest/reference/#operation/batchAnalyse>`__

        :param batch_request: It could be a batch request object, a raw batch request payload or only a batch
            request ID.
        """
        return self._call_job(batch_request, "analyse")

    def start_job(self, batch_request: BatchRequestType) -> Json:
        """Starts running a batch job

        `Batch API reference <https://docs.sentinel-hub.com/api/latest/reference/#operation/batchStartProcessRequest>`__

        :param batch_request: It could be a batch request object, a raw batch request payload or only a batch
            request ID.
        """
        return self._call_job(batch_request, "start")

    def cancel_job(self, batch_request: BatchRequestType) -> Json:
        """Cancels a batch job

        `Batch API reference
        <https://docs.sentinel-hub.com/api/latest/reference/#operation/batchCancelProcessRequest>`__

        :param batch_request: It could be a batch request object, a raw batch request payload or only a batch
            request ID.
        """
        return self._call_job(batch_request, "cancel")

    def restart_job(self, batch_request: BatchRequestType) -> Json:
        """Restarts only those parts of a job that failed

        `Batch API reference
        <https://docs.sentinel-hub.com/api/latest/reference/#operation/batchRestartPartialProcessRequest>`__

        :param batch_request: It could be a batch request object, a raw batch request payload or only a batch
            request ID.
        """
        return self._call_job(batch_request, "restartpartial")

    def iter_tiles(
        self, batch_request: BatchRequestType, status: Union[None, "BatchTileStatus", str] = None, **kwargs: Any
    ) -> SentinelHubFeatureIterator:
        """Iterate over info about batch request tiles

        `Batch API reference <https://docs.sentinel-hub.com/api/latest/reference/#operation/getAllBatchProcessTiles>`__

        :param batch_request: It could be a batch request object, a raw batch request payload or only a batch
            request ID.
        :param status: A filter to obtain only tiles with a certain status
        :param kwargs: Any additional parameters to include in a request query
        :return: An iterator over information about each tile
        """
        request_id = self._parse_request_id(batch_request)
        if isinstance(status, BatchTileStatus):
            status = status.value

        return SentinelHubFeatureIterator(
            client=self.client,
            url=self._get_tiles_url(request_id),
            params={"status": status, **kwargs},
            exception_message="No tiles found, please run analysis on batch request before calling this method",
        )

    def get_tile(self, batch_request: BatchRequestType, tile_id: Optional[int]) -> JsonDict:
        """Provides information about a single batch request tile

        `Batch API reference <https://docs.sentinel-hub.com/api/latest/reference/#operation/getBatchTileById>`__

        :param batch_request: It could be a batch request object, a raw batch request payload or only a batch
            request ID.
        :param tile_id: An ID of a tile
        :return: Information about a tile
        """
        request_id = self._parse_request_id(batch_request)
        url = self._get_tiles_url(request_id, tile_id=tile_id)
        return self.client.get_json_dict(url, use_session=True)

    def reprocess_tile(self, batch_request: BatchRequestType, tile_id: Optional[int]) -> Json:
        """Reprocess a single failed tile

        `Batch API reference <https://docs.sentinel-hub.com/api/latest/reference/#operation/restartBatchTileById>`__

        :param batch_request: It could be a batch request object, a raw batch request payload or only a batch
            request ID.
        :param tile_id: An ID of a tile
        """
        return self._call_job(batch_request, f"tiles/{tile_id}/restart")

    def iter_collections(self, search: Optional[str] = None, **kwargs: Any) -> SentinelHubFeatureIterator:
        """Iterate over batch collections

        `Batch API reference <https://docs.sentinel-hub.com/api/latest/reference/#operation/getAllBatchCollections>`__

        :param search: A search query to filter collections
        :param kwargs: Any additional parameters to include in a request query
        :return: An iterator over existing batch collections
        """
        return SentinelHubFeatureIterator(
            client=self.client,
            url=self._get_collections_url(),
            params={"search": search, **kwargs},
            exception_message="Failed to obtain information about available Batch collections",
        )

    def get_collection(self, collection_id: str) -> JsonDict:
        """Get batch collection by its id

        `Batch API reference
        <https://docs.sentinel-hub.com/api/latest/reference/#operation/getSingleBatchCollectionById>`__

        :param collection_id: A batch collection id
        :return: A dictionary of the collection parameters
        """
        url = self._get_collections_url(collection_id)
        return self.client.get_json_dict(url=url, use_session=True, extract_key="data")

    def create_collection(self, collection: Union["BatchCollection", dict]) -> JsonDict:
        """Create a new batch collection

        `Batch API reference <https://docs.sentinel-hub.com/api/latest/reference/#operation/createNewBatchCollection>`__

        :param collection: Batch collection definition
        :return: A dictionary of a newly created collection
        """
        collection_payload = self._parse_collection_to_dict(collection)
        url = self._get_collections_url()
        return self.client.get_json_dict(url=url, post_values=collection_payload, use_session=True, extract_key="data")

    def update_collection(self, collection: Union["BatchCollection", dict]) -> Json:
        """Update an existing batch collection

        `Batch API reference <https://docs.sentinel-hub.com/api/latest/reference/#operation/updateBatchCollection>`__

        :param collection: Batch collection definition
        """
        collection_id = self._parse_collection_id(collection)
        return self.client.get_json(
            url=self._get_collections_url(collection_id),
            post_values=self._parse_collection_to_dict(collection),
            request_type=RequestType.PUT,
            use_session=True,
        )

    def delete_collection(self, collection: BatchCollectionType) -> Json:
        """Delete an existing batch collection

        `Batch API reference <https://docs.sentinel-hub.com/api/latest/reference/#operation/deleteBatchCollection>`__

        :param collection: Batch collection id or object
        """
        collection_id = self._parse_collection_id(collection)
        return self.client.get_json(
            url=self._get_collections_url(collection_id), request_type=RequestType.DELETE, use_session=True
        )

    def _call_job(self, batch_request: BatchRequestType, endpoint_name: str) -> Json:
        """Makes a POST request to the service that triggers a processing job"""
        request_id = self._parse_request_id(batch_request)
        job_url = f"{self._get_process_url(request_id)}/{endpoint_name}"
        return self.client.get_json(url=job_url, request_type=RequestType.POST, use_session=True)

    def _get_process_url(self, request_id: Optional[str] = None) -> str:
        """Creates a URL for process endpoint"""
        url = f"{self.service_url}/process"
        if request_id:
            return f"{url}/{request_id}"
        return url

    def _get_tiles_url(self, request_id: str, tile_id: Union[None, str, int] = None) -> str:
        """Creates a URL for tiles endpoint"""
        url = f"{self._get_process_url(request_id)}/tiles"
        if tile_id:
            return f"{url}/{tile_id}"
        return url

    def _get_tiling_grids_url(self, grid_id: Union[None, str, int] = None) -> str:
        """Creates a URL for tiling grids endpoint"""
        url = f"{self.service_url}/tilinggrids"
        if grid_id:
            return f"{url}/{grid_id}"
        return url

    def _get_collections_url(self, collection_id: Optional[str] = None) -> str:
        """Creates a URL for batch collections endpoint"""
        url = f"{self.service_url}/collections"
        if collection_id:
            return f"{url}/{collection_id}"
        return url

    @staticmethod
    def _parse_request_id(data: BatchRequestType) -> str:
        """Parses batch request id from multiple possible inputs"""
        if isinstance(data, BatchRequest):
            return data.request_id
        if isinstance(data, dict):
            return data["id"]
        if isinstance(data, str):
            return data
        raise ValueError(f"Expected a BatchRequest, dictionary or a string, got {data}.")

    @staticmethod
    def _parse_collection_id(data: BatchCollectionType) -> Optional[str]:
        """Parses batch collection id from multiple possible inputs"""
        if isinstance(data, (BatchCollection, DataCollection)):
            return data.collection_id
        if isinstance(data, dict):
            return data["id"]
        if isinstance(data, str):
            return data
        raise ValueError(f"Expected a BatchCollection dataclass, dictionary or a string, got {data}.")

    @staticmethod
    def _parse_collection_to_dict(data: Union["BatchCollection", dict]) -> dict:
        """Constructs a dictionary from given object"""
        if isinstance(data, BatchCollection):
            return data.to_dict()  # type: ignore[attr-defined]
        if isinstance(data, dict):
            return data
        raise ValueError(f"Expected either a BatchCollection or a dict, got {data}.")


class BatchRequestStatus(Enum):
    """An enum class with all possible batch request statuses"""

    CREATED = "CREATED"
    ANALYSING = "ANALYSING"
    ANALYSIS_DONE = "ANALYSIS_DONE"
    PROCESSING = "PROCESSING"
    DONE = "DONE"
    FAILED = "FAILED"
    PARTIAL = "PARTIAL"
    CANCELED = "CANCELED"


class BatchTileStatus(Enum):
    """An enum class with all possible batch tile statuses"""

    PENDING = "PENDING"
    SCHEDULED = "SCHEDULED"
    QUEUED = "QUEUED"
    PROCESSING = "PROCESSING"
    PROCESSED = "PROCESSED"
    FAILED = "FAILED"


class BatchUserAction(Enum):
    """An enum class with all possible batch user actions"""

    START = "START"
    ANALYSE = "ANALYSE"
    NONE = "NONE"
    CANCEL = "CANCEL"


@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.INCLUDE)
@dataclass
class BatchRequest:
    """A dataclass object that holds information about a batch request"""

    request_id: str = field(metadata=dataclass_config(field_name="id"))
    process_request: dict
    tile_count: int
    status: BatchRequestStatus = field(metadata=enum_config(BatchRequestStatus))
    other_data: CatchAll
    user_id: Optional[str] = None
    created: Optional[dt.datetime] = field(metadata=datetime_config, default=None)
    tiling_grid: dict = field(default_factory=dict)
    output: dict = field(default_factory=dict)
    bucket_name: Optional[str] = None
    description: Optional[str] = None
    value_estimate: Optional[float] = None
    tile_width_px: Optional[int] = None
    tile_height_px: Optional[int] = None
    user_action: Optional[BatchUserAction] = field(metadata=enum_config(BatchUserAction), default=None)
    user_action_updated: Optional[str] = field(metadata=datetime_config, default=None)
    error: Optional[str] = None

    _REPR_PARAM_NAMES = [
        "request_id",
        "description",
        "bucket_name",
        "created",
        "status",
        "user_action",
        "value_estimate",
        "tile_count",
    ]

    def __repr__(self) -> str:
        """A representation that shows the basic parameters of a batch job"""
        repr_params = {name: getattr(self, name) for name in self._REPR_PARAM_NAMES if getattr(self, name) is not None}
        repr_params_str = "\n  ".join(f"{name}={value}" for name, value in repr_params.items())
        return f"{self.__class__.__name__}(\n  {repr_params_str}\n  ...\n)"

    @property
    def evalscript(self) -> str:
        """Provides an evalscript used by a batch request

        :return: An evalscript
        """
        return self.process_request["evalscript"]

    @property
    def bbox(self) -> BBox:
        """Provides a bounding box used by a batch request

        :return: An area bounding box together with CRS
        :raises: ValueError
        """
        bbox, _, crs = self._parse_bounds_payload()
        if bbox is None:
            raise ValueError("Bounding box is not defined for this batch request")
        return BBox(bbox, crs)

    @property
    def geometry(self) -> Geometry:
        """Provides a geometry used by a batch request

        :return: An area geometry together with CRS
        :raises: ValueError
        """
        _, geometry, crs = self._parse_bounds_payload()
        if geometry is None:
            raise ValueError("Geometry is not defined for this batch request")
        return Geometry(geometry, crs)

    def raise_for_status(
        self,
        status: Union[str, BatchRequestStatus, Iterable[Union[str, BatchRequestStatus]]] = BatchRequestStatus.FAILED,
    ) -> None:
        """Raises an error in case batch request has a given status

        :param status: One or more status codes on which to raise an error. The default is `'FAILED'`.
        :raises: RuntimeError
        """
        if isinstance(status, (str, BatchRequestStatus)):
            status = [status]
        status_list = [BatchRequestStatus(_status) for _status in status]

        if self.status in status_list:
            formatted_error_message = f' and error message: "{self.error}"' if self.error else ""
            raise RuntimeError(
                f"Raised for batch request {self.request_id} with status {self.status.value}{formatted_error_message}"
            )

    def _parse_bounds_payload(self) -> Tuple[Optional[List[float]], Optional[list], CRS]:
        """Parses bbox, geometry and crs from batch request payload. If bbox or geometry don't exist it returns None
        instead.
        """
        bounds_definition = self.process_request["input"]["bounds"]
        crs = CRS(bounds_definition["properties"]["crs"].rsplit("/", 1)[-1])

        return bounds_definition.get("bbox"), bounds_definition.get("geometry"), crs


@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.INCLUDE)
@dataclass
class BatchCollectionBatchData:
    """Dataclass to hold batch collection batchData part of the payload"""

    other_data: CatchAll
    tiling_grid_id: Optional[int] = None


@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.INCLUDE)
@dataclass
class BatchCollectionAdditionalData:
    """Dataclass to hold batch collection additionalData part of the payload"""

    other_data: CatchAll
    bands: Optional[Dict[str, Any]] = None


class BatchCollection(BaseCollection):
    """Dataclass for batch collections"""

    batch_data: Optional[BatchCollectionBatchData] = None
    additional_data: Optional[BatchCollectionAdditionalData] = None


_MIN_SLEEP_TIME = 60
_DEFAULT_SLEEP_TIME = 120
_MIN_ANALYSIS_SLEEP_TIME = 5
_DEFAULT_ANALYSIS_SLEEP_TIME = 10


def monitor_batch_job(
    batch_request: BatchRequestType,
    config: Optional[SHConfig],
    sleep_time: int = _DEFAULT_SLEEP_TIME,
    analysis_sleep_time: int = _DEFAULT_ANALYSIS_SLEEP_TIME,
) -> DefaultDict[BatchTileStatus, List[dict]]:
    """A utility function that keeps checking for number of processed tiles until the given batch request finishes.
    During the process it shows a progress bar and at the end it reports information about finished and failed tiles.

    Notes:

      - Before calling this function make sure to start a batch job by calling `SentinelHubBatch.start_job` method. In
        case a batch job is still being analysed this function will wait until the analysis ends.
      - This function will be continuously collecting tile information from Sentinel Hub service. To avoid making too
        many requests please make sure to adjust `sleep_time` parameter according to the size of your job. Larger jobs
        don't need very frequent tile status updates.
      - Some information about the progress of this function is reported to logging level INFO.

    :param batch_request: An object with information about a batch request. Alternatively, it could only be a batch
        request id or a payload.
    :param config: A configuration object with required parameters `sh_client_id`, `sh_client_secret`, and
        `sh_auth_base_url` which is used for authentication and `sh_base_url` which defines the service deployment
        where Batch API will be called.
    :param sleep_time: Number of seconds to sleep between consecutive progress bar updates.
    :param analysis_sleep_time: Number of seconds between consecutive status updates during analysis phase.
    :return: A dictionary mapping a tile status to a list of tile payloads.
    """
    if sleep_time < _MIN_SLEEP_TIME:
        raise ValueError(f"To avoid making too many service requests please set sleep_time>={_MIN_SLEEP_TIME}")

    batch_request = monitor_batch_analysis(batch_request, config, sleep_time=analysis_sleep_time)
    if batch_request.status is BatchRequestStatus.PROCESSING:
        LOGGER.info("Batch job is running")

    batch_client = SentinelHubBatch(config=config)

    tiles_per_status = _get_batch_tiles_per_status(batch_request, batch_client)
    success_count = len(tiles_per_status[BatchTileStatus.PROCESSED])
    finished_count = success_count + len(tiles_per_status[BatchTileStatus.FAILED])

    with tqdm(total=batch_request.tile_count, initial=finished_count, desc="Progress rate") as progress_bar, tqdm(
        total=finished_count, initial=success_count, desc="Success rate"
    ) as success_bar:
        while finished_count < batch_request.tile_count:
            time.sleep(sleep_time)

            tiles_per_status = _get_batch_tiles_per_status(batch_request, batch_client)
            new_success_count = len(tiles_per_status[BatchTileStatus.PROCESSED])
            new_finished_count = new_success_count + len(tiles_per_status[BatchTileStatus.FAILED])

            progress_bar.update(new_finished_count - finished_count)
            if new_finished_count != finished_count:
                success_bar.total = new_finished_count
                success_bar.refresh()
            success_bar.update(new_success_count - success_count)

            finished_count = new_finished_count
            success_count = new_success_count

    failed_tiles_num = finished_count - success_count
    if failed_tiles_num:
        LOGGER.info("Batch job failed for %d tiles", failed_tiles_num)
    return tiles_per_status


def monitor_batch_analysis(
    batch_request: BatchRequestType, config: Optional[SHConfig], sleep_time: int = _DEFAULT_ANALYSIS_SLEEP_TIME
) -> BatchRequest:
    """A utility function that is waiting until analysis phase of a batch job finishes and regularly checks its status.
    In case analysis phase failed it raises an error at the end.

    :param batch_request: An object with information about a batch request. Alternatively, it could only be a batch
        request id or a payload.
    :param config: A configuration object with required parameters `sh_client_id`, `sh_client_secret`, and
        `sh_auth_base_url` which is used for authentication and `sh_base_url` which defines the service deployment
        where Batch API will be called.
    :param sleep_time: Number of seconds between consecutive status updates during analysis phase.
    :return: Batch request info
    """
    if sleep_time < _MIN_ANALYSIS_SLEEP_TIME:
        raise ValueError(
            f"To avoid making too many service requests please set analysis sleep time >={_MIN_ANALYSIS_SLEEP_TIME}"
        )

    batch_client = SentinelHubBatch(config=config)

    batch_request = batch_client.get_request(batch_request)
    while batch_request.status in [BatchRequestStatus.CREATED, BatchRequestStatus.ANALYSING]:
        LOGGER.info("Batch job has a status %s, sleeping for %d seconds", batch_request.status.value, sleep_time)
        time.sleep(sleep_time)
        batch_request = batch_client.get_request(batch_request)

    batch_request.raise_for_status(status=[BatchRequestStatus.FAILED, BatchRequestStatus.CANCELED])
    return batch_request


def _get_batch_tiles_per_status(
    batch_request: BatchRequestType, batch_client: SentinelHubBatch
) -> DefaultDict[BatchTileStatus, List[dict]]:
    """A helper function that queries information about batch tiles and returns information about tiles, grouped by
    tile status.

    :return: A dictionary mapping a tile status to a list of tile payloads.
    """
    tiles_per_status = defaultdict(list)

    for tile in batch_client.iter_tiles(batch_request):
        status = BatchTileStatus(tile["status"])
        tiles_per_status[status].append(tile)

    return tiles_per_status



"""
Module implementing error handlers which can occur during download procedure
"""
import logging
import sys
import time
from typing import Callable, Optional, TypeVar

import requests

from ..config import SHConfig
from ..decoding import decode_sentinelhub_err_msg
from ..exceptions import DownloadFailedException
from .request import DownloadRequest

if sys.version_info < (3, 8):
    from typing_extensions import Protocol
else:
    from typing import Protocol  # pylint: disable=ungrouped-imports


class _HasConfig(Protocol):
    """Interface of objects with a config."""

    config: SHConfig


Self = TypeVar("Self")
SelfWithConfig = TypeVar("SelfWithConfig", bound=_HasConfig)
T = TypeVar("T")


LOGGER = logging.getLogger(__name__)


def fail_user_errors(download_func: Callable[[Self, DownloadRequest], T]) -> Callable[[Self, DownloadRequest], T]:
    """Decorator function for handling user errors"""

    def new_download_func(self: Self, request: DownloadRequest) -> T:
        try:
            return download_func(self, request)
        except requests.HTTPError as exception:
            if (
                exception.response.status_code < requests.status_codes.codes.INTERNAL_SERVER_ERROR
                and exception.response.status_code != requests.status_codes.codes.TOO_MANY_REQUESTS
            ):
                raise DownloadFailedException(
                    _create_download_failed_message(exception, request.url), request_exception=exception
                ) from exception
            raise exception from exception

    return new_download_func


def retry_temporary_errors(
    download_func: Callable[[SelfWithConfig, DownloadRequest], T]
) -> Callable[[SelfWithConfig, DownloadRequest], T]:
    """Decorator function for handling server and connection errors"""
    backoff_coefficient = 3

    def new_download_func(self: SelfWithConfig, request: DownloadRequest) -> T:
        download_attempts = self.config.max_download_attempts
        sleep_time = self.config.download_sleep_time

        for attempt_num in range(download_attempts):
            try:
                return download_func(self, request)
            except requests.RequestException as exception:
                if not (
                    _is_temporary_problem(exception)
                    or (
                        isinstance(exception, requests.HTTPError)
                        and exception.response.status_code >= requests.status_codes.codes.INTERNAL_SERVER_ERROR
                    )
                ):
                    raise exception from exception

                if attempt_num == download_attempts - 1:
                    message = _create_download_failed_message(exception, request.url)
                    raise DownloadFailedException(message, request_exception=exception) from exception

                LOGGER.debug(
                    "Download attempt failed: %s\n%d attempts left, will retry in %ds",
                    exception,
                    download_attempts - attempt_num - 1,
                    sleep_time,
                )
                time.sleep(sleep_time)
                sleep_time *= backoff_coefficient

        raise DownloadFailedException(
            "No download attempts available - configuration parameter max_download_attempts should be greater than 0"
        )

    return new_download_func


def fail_missing_file(download_func: Callable[[Self, DownloadRequest], T]) -> Callable[[Self, DownloadRequest], T]:
    """A decorator for raising an error if a file is missing"""

    def new_download_func(self: Self, request: DownloadRequest) -> T:
        try:
            return download_func(self, request)
        except requests.HTTPError as exception:
            if exception.response.status_code == requests.status_codes.codes.NOT_FOUND:
                raise DownloadFailedException(
                    f"File in location {request.url} is missing", request_exception=exception
                ) from exception

            raise exception from exception

    return new_download_func


def _is_temporary_problem(exception: Exception) -> bool:
    """Checks if the obtained exception is temporary and if download attempt should be repeated

    :param exception: Exception raised during download
    :return: `True` if exception is temporary and `False` otherwise
    """
    return isinstance(exception, (requests.ConnectionError, requests.Timeout, requests.exceptions.ChunkedEncodingError))


def _create_download_failed_message(exception: Exception, url: Optional[str]) -> str:
    """Creates message describing why download has failed

    :param exception: Exception raised during download
    :param url: A URL from where download was attempted
    :return: Error message
    """
    message = f"Failed to download from:\n{url}\nwith {exception.__class__.__name__}:\n{exception}"

    if _is_temporary_problem(exception):
        if isinstance(exception, requests.ConnectionError):
            message += "\nPlease check your internet connection and try again."
        else:
            message += (
                "\nThere might be a problem in connection or the server failed to process "
                "your request. Please try again."
            )
    elif isinstance(exception, requests.HTTPError):
        server_message = decode_sentinelhub_err_msg(exception.response)
        message += f'\nServer response: "{server_message}"'

    return message



"""
Module defining custom package exceptions
"""
import warnings
from typing import Optional

import requests


class BaseSentinelHubException(Exception):
    """A base package exception"""


class DownloadFailedException(BaseSentinelHubException):
    """General exception which is raised whenever download fails"""

    def __init__(self, msg: str, *, request_exception: Optional[requests.RequestException] = None) -> None:
        super().__init__(msg)
        self.request_exception = request_exception


class OutOfRequestsException(DownloadFailedException):
    """An exception raised whenever download cannot be done because user has run out of requests or processing units"""


class AwsDownloadFailedException(DownloadFailedException):
    """This exception is raised when download fails because of a missing file in AWS"""


class ImageDecodingError(BaseSentinelHubException):
    """This exception is raised when downloaded image is not properly decoded"""


class MissingDataInRequestException(BaseSentinelHubException):
    """This exception is raised when an iteration is performed on a request without data."""


class HashedNameCollisionException(BaseSentinelHubException):
    """This exception is raised when two different requests are assigned the same hashed name."""


class SHDeprecationWarning(DeprecationWarning):
    """A custom deprecation warning for sentinelhub-py package"""


class SHUserWarning(UserWarning):
    """A custom user warning for sentinelhub-py package"""


class SHRuntimeWarning(RuntimeWarning):
    """A custom runtime warning for sentinelhub-py package"""


class SHRateLimitWarning(SHRuntimeWarning):
    """A custom runtime warning in case user hit the rate limit for downloads"""


warnings.simplefilter("default", SHDeprecationWarning)
warnings.simplefilter("default", SHUserWarning)
warnings.simplefilter("always", SHRuntimeWarning)
warnings.simplefilter("always", SHRateLimitWarning)



"""
Module for working with Sentinel Hub OGC services
`Sentinel Hub OGC services <https://www.sentinel-hub.com/develop/api/ogc/standard-parameters/>`__.
"""

import datetime
import logging
from base64 import b64encode
from typing import Any, Dict, List, Optional, Tuple, Union
from urllib.parse import urlencode

from ..base import DataRequest
from ..config import SHConfig
from ..constants import CRS, CustomUrlParam, MimeType, ResamplingType, ServiceType, SHConstants
from ..data_collections import DataCollection
from ..download import DownloadRequest, SentinelHubDownloadClient
from ..geo_utils import get_image_dimension
from ..geometry import BBox, Geometry
from ..time_utils import RawTimeIntervalType, RawTimeType, filter_times, parse_time_interval, serialize_time
from .wfs import WebFeatureService

LOGGER = logging.getLogger(__name__)


class OgcRequest(DataRequest):
    """The base class for OGC-type requests (WMS and WCS) where all common parameters are defined"""

    def __init__(
        self,
        layer: str,
        bbox: BBox,
        *,
        data_collection: DataCollection,
        time: Union[RawTimeType, RawTimeIntervalType] = "latest",
        service_type: Optional[ServiceType] = None,
        size_x: Union[None, str, int] = None,
        size_y: Union[None, str, int] = None,
        maxcc: float = 1.0,
        image_format: MimeType = MimeType.PNG,
        custom_url_params: Optional[Dict[CustomUrlParam, Any]] = None,
        time_difference: datetime.timedelta = datetime.timedelta(seconds=-1),
        **kwargs: Any,
    ):
        """
        :param layer: An ID of a layer configured in Sentinel Hub Dashboard. It has to be configured for the same
            instance ID which will be used for this request. The satellite collection of the layer in Dashboard
            must also match the one given by `data_collection` parameter
        :param bbox: Bounding box of the requested image. Coordinates must be in the specified coordinate reference
            system.
        :param data_collection: A collection of requested satellite data. It has to be the same as defined in
            Sentinel Hub Dashboard for the given layer.
        :param time: time or time range for which to return the results, in ISO8601 format
            (year-month-date, for example: ``2016-01-01``, or year-month-dateThours:minutes:seconds format,
            i.e. ``2016-01-01T16:31:21``). When a single time is specified the request will return data for that
            specific date, if it exists. If a time range is specified the result is a list of all scenes between the
            specified dates conforming to the cloud coverage criteria. Most recent acquisition being first in the list.
            For the latest acquisition use ``latest``. Examples: ``latest``, ``'2016-01-01'``, or
            ``('2016-01-01', ' 2016-01-31')``
        :param service_type: type of OGC service (WMS or WCS)
        :param size_x: number of pixels in x or resolution in x (i.e. ``512`` or ``10m``)
        :param size_y: number of pixels in x or resolution in y (i.e. ``512`` or ``10m``)
        :param maxcc: maximum accepted cloud coverage of an image. Float between 0.0 and 1.0. Default is ``1.0``.
        :param image_format: format of the returned image by the Sentinel Hub's WMS getMap service. Default is PNG, but
                            in some cases 32-bit TIFF is required, i.e. if requesting unprocessed raw bands.
                            Default is ``constants.MimeType.PNG``.
        :param custom_url_params: A dictionary of CustomUrlParameters and their values supported by Sentinel Hub's WMS
            and WCS services. All available parameters are described in
            `documentation <https://www.sentinel-hub.com/develop/api/ogc/custom-parameters/>`__. Note: in case of
            `CustomUrlParam.EVALSCRIPT` the dictionary value must be a string of Javascript code that is not
            encoded into base64.
        :param time_difference: The time difference below which dates are deemed equal. That is, if for the given set
            of OGC parameters the images are available at datetimes `d1<=d2<=...<=dn` then only those with
            `dk-dj>time_difference` will be considered. The default time difference is negative (`-1s`), meaning
            that all dates are considered by default.
        :param data_folder: location of the directory where the fetched data will be saved.
        :param config: A custom instance of config class to override parameters from the saved configuration.
        """
        self.layer = layer
        self.bbox = bbox
        self.time = time
        self.data_collection = data_collection
        self.maxcc = maxcc
        self.image_format = MimeType(image_format)
        self.service_type = service_type
        self.size_x = size_x
        self.size_y = size_y
        self.custom_url_params = custom_url_params
        self.time_difference = time_difference

        self._check_custom_url_parameters()

        self.wfs_iterator: Optional[WebFeatureService] = None

        super().__init__(SentinelHubDownloadClient, **kwargs)

    def _check_custom_url_parameters(self) -> None:
        """Checks if custom url parameters are valid parameters.

        Throws ValueError if the provided parameter is not a valid parameter.
        """
        if self.custom_url_params is None:
            return

        for param in self.custom_url_params:
            if param not in CustomUrlParam:
                raise ValueError(f"Parameter {param} is not a valid custom url parameter. Please check and fix.")

        if self.service_type is ServiceType.FIS and CustomUrlParam.GEOMETRY in self.custom_url_params:
            raise ValueError(f"{CustomUrlParam.GEOMETRY} should not be a custom url parameter of a FIS request")

    def create_request(self, reset_wfs_iterator: bool = False) -> None:
        """Set download requests

        Create a list of DownloadRequests for all Sentinel-2 acquisitions within request's time interval and
        acceptable cloud coverage.

        :param reset_wfs_iterator: When re-running the method this flag is used to reset/keep existing ``wfs_iterator``
            (i.e. instance of ``WebFeatureService`` class). If the iterator is not reset you don't have to repeat a
            service call but tiles and dates will stay the same.
        """
        if reset_wfs_iterator:
            self.wfs_iterator = None

        ogc_service = OgcImageService(config=self.config)
        self.download_list = ogc_service.get_request(self)
        self.wfs_iterator = ogc_service.get_wfs_iterator()

    def get_dates(self) -> List[Optional[datetime.datetime]]:
        """Get list of dates

        List of all available Sentinel-2 acquisitions for given bbox with max cloud coverage and the specified
        time interval. When a single time is specified the request will return that specific date, if it exists.
        If a time range is specified the result is a list of all scenes between the specified dates conforming to
        the cloud coverage criteria. Most recent acquisition being first in the list.

        :return: list of all available Sentinel-2 acquisition times within request's time interval and
            acceptable cloud coverage.
        """
        return OgcImageService(config=self.config).get_dates(self)

    def get_tiles(self) -> Optional[WebFeatureService]:
        """Returns iterator over info about all satellite tiles used for the OgcRequest

        :return: Iterator of dictionaries containing info about all satellite tiles used in the request. In case of
            `DataCollection.DEM` it returns None.
        """
        return self.wfs_iterator


class WmsRequest(OgcRequest):
    """Web Map Service request class

    Creates an instance of Sentinel Hub WMS (Web Map Service) GetMap request,
    which provides access to Sentinel-2's unprocessed bands (B01, B02, ..., B08, B8A, ..., B12)
    or processed products such as true color imagery, NDVI, etc. The only difference is that in
    the case of WMS request the user specifies the desired image size instead of its resolution.

    It is required to specify at least one of `width` and `height` parameters. If only one of them is specified the
    other one will be calculated to best fit the bounding box ratio. If both of them are specified they will be used
    no matter the bounding box ratio.

    For more info check `WMS documentation <https://www.sentinel-hub.com/develop/api/ogc/standard-parameters/wms/>`__.
    """

    def __init__(self, *, width: Optional[int] = None, height: Optional[int] = None, **kwargs: Any):
        """
        :param width: width (number of columns) of the returned image (array)
        :param height: height (number of rows) of the returned image (array)
        :param layer: An ID of a layer configured in Sentinel Hub Dashboard. It has to be configured for the same
            instance ID which will be used for this request. The satellite collection of the layer in Dashboard
            must also match the one given by `data_collection` parameter
        :param bbox: Bounding box of the requested image. Coordinates must be in the specified coordinate reference
            system.
        :param time: time or time range for which to return the results, in ISO8601 format
            (year-month-date, for example: ``2016-01-01``, or year-month-dateThours:minutes:seconds format,
            i.e. ``2016-01-01T16:31:21``). When a single time is specified the request will return data for that
            specific date, if it exists. If a time range is specified the result is a list of all scenes between the
            specified dates conforming to the cloud coverage criteria. Most recent acquisition being first in the list.
            For the latest acquisition use ``latest``. Examples: ``latest``, ``'2016-01-01'``, or
            ``('2016-01-01', ' 2016-01-31')``
            (datetime.datetime, datetime.datetime)
        :param data_collection: A collection of requested satellite data. It has to be the same as defined in
            Sentinel Hub Dashboard for the given layer. Default is Sentinel-2 L1C.
        :param size_x: number of pixels in x or resolution in x (i.e. ``512`` or ``10m``)
        :param size_y: number of pixels in x or resolution in y (i.e. ``512`` or ``10m``)
        :param maxcc: maximum accepted cloud coverage of an image. Float between 0.0 and 1.0. Default is ``1.0``.
        :param image_format: format of the returned image by the Sentinel Hub's WMS getMap service. Default is PNG, but
                            in some cases 32-bit TIFF is required, i.e. if requesting unprocessed raw bands.
                            Default is ``constants.MimeType.PNG``.
        :param custom_url_params: A dictionary of CustomUrlParameters and their values supported by Sentinel Hub's WMS
            and WCS services. All available parameters are described in
            `documentation <https://www.sentinel-hub.com/develop/api/ogc/custom-parameters/>`__. Note: in case of
            `CustomUrlParam.EVALSCRIPT` the dictionary value must be a string of Javascript code that is not
            encoded into base64.
        :param time_difference: The time difference below which dates are deemed equal. That is, if for the given set
            of OGC parameters the images are available at datetimes `d1<=d2<=...<=dn` then only those with
            `dk-dj>time_difference` will be considered. The default time difference is negative (`-1s`), meaning
            that all dates are considered by default.
        :param data_folder: location of the directory where the fetched data will be saved.
        :param config: A custom instance of config class to override parameters from the saved configuration.
        """
        super().__init__(service_type=ServiceType.WMS, size_x=width, size_y=height, **kwargs)


class WcsRequest(OgcRequest):
    """Web Coverage Service request class

    Creates an instance of Sentinel Hub WCS (Web Coverage Service) GetCoverage request,
    which provides access to Sentinel-2's unprocessed bands (B01, B02, ..., B08, B8A, ..., B12)
    or processed products such as true color imagery, NDVI, etc., as the WMS service. The
    only difference is that in the case of WCS request the user specifies the desired
    resolution of the image instead of its size.

    For more info check `WCS documentation <https://www.sentinel-hub.com/develop/api/ogc/standard-parameters/wcs/>`__.
    """

    def __init__(self, *, resx: str = "10m", resy: str = "10m", **kwargs: Any):
        """
        :param resx: resolution in x (resolution of a column) given in meters in the format (examples ``10m``,
            ``20m``, ...). Default is ``10m``, which is the best native resolution of some Sentinel-2 bands.
        :param resy: resolution in y (resolution of a row) given in meters in the format
            (examples ``10m``, ``20m``, ...). Default is ``10m``, which is the best native resolution of some
            Sentinel-2 bands.
        :param layer: An ID of a layer configured in Sentinel Hub Dashboard. It has to be configured for the same
            instance ID which will be used for this request. The satellite collection of the layer in Dashboard
            must also match the one given by `data_collection` parameter
        :param bbox: Bounding box of the requested image. Coordinates must be in the specified coordinate reference
            system.
        :param time: time or time range for which to return the results, in ISO8601 format
            (year-month-date, for example: ``2016-01-01``, or year-month-dateThours:minutes:seconds format,
            i.e. ``2016-01-01T16:31:21``). When a single time is specified the request will return data for that
            specific date, if it exists. If a time range is specified the result is a list of all scenes between the
            specified dates conforming to the cloud coverage criteria. Most recent acquisition being first in the list.
            For the latest acquisition use ``latest``. Examples: ``latest``, ``'2016-01-01'``, or
            ``('2016-01-01', ' 2016-01-31')``
            (datetime.datetime, datetime.datetime)
        :param data_collection: A collection of requested satellite data. It has to be the same as defined in Sentinel
            Hub Dashboard for the given layer. Default is Sentinel-2 L1C.
        :param size_x: number of pixels in x or resolution in x (i.e. ``512`` or ``10m``)
        :param size_y: number of pixels in x or resolution in y (i.e. ``512`` or ``10m``)
        :param maxcc: maximum accepted cloud coverage of an image. Float between 0.0 and 1.0. Default is ``1.0``.
        :param image_format: format of the returned image by the Sentinel Hub's WMS getMap service. Default is PNG, but
                            in some cases 32-bit TIFF is required, i.e. if requesting unprocessed raw bands.
                            Default is ``constants.MimeType.PNG``.
        :param custom_url_params: A dictionary of CustomUrlParameters and their values supported by Sentinel Hub's WMS
            and WCS services. All available parameters are described in
            `documentation <https://www.sentinel-hub.com/develop/api/ogc/custom-parameters/>`__. Note: in case of
            `CustomUrlParam.EVALSCRIPT` the dictionary value must be a string of Javascript code that is not
            encoded into base64.
        :param time_difference: The time difference below which dates are deemed equal. That is, if for the given set
            of OGC parameters the images are available at datetimes `d1<=d2<=...<=dn` then only those with
            `dk-dj>time_difference` will be considered. The default time difference is negative (`-1s`), meaning
            that all dates are considered by default.
        :param data_folder: location of the directory where the fetched data will be saved.
        :param config: A custom instance of config class to override parameters from the saved configuration.
        """
        super().__init__(service_type=ServiceType.WCS, size_x=resx, size_y=resy, **kwargs)


class OgcImageService:
    """Sentinel Hub OGC services class for providing image data

    Intermediate layer between QGC-type requests (WmsRequest and WcsRequest) and the Sentinel Hub OGC (WMS and WCS)
    services.
    """

    def __init__(self, config: Optional[SHConfig] = None):
        """
        :param config: A custom instance of config class to override parameters from the saved configuration.
        """
        self.config = config or SHConfig()
        self.config.raise_for_missing_instance_id()

        self._base_url = self.config.get_sh_ogc_url()
        self.wfs_iterator: Optional[WebFeatureService] = None

    def get_request(self, request: OgcRequest) -> List[DownloadRequest]:
        """Get download requests

        Create a list of DownloadRequests for all Sentinel-2 acquisitions within request's time interval and
        acceptable cloud coverage.

        :param request: QGC-type request with specified bounding box, time interval, and cloud coverage for specific
                        product.
        :return: list of DownloadRequests
        """
        size_x, size_y = self.get_image_dimensions(request)
        return [
            DownloadRequest(
                url=self.get_url(request=request, date=date, size_x=size_x, size_y=size_y),
                data_type=request.image_format,
                headers=SHConstants.HEADERS,
            )
            for date in self.get_dates(request)
        ]

    def get_url(
        self,
        request: OgcRequest,
        *,
        date: Optional[datetime.datetime] = None,
        size_x: Union[None, str, int] = None,
        size_y: Union[None, str, int] = None,
        geometry: Union[None, BBox, Geometry] = None,
    ) -> str:
        """Returns url to Sentinel Hub's OGC service for the product specified by the OgcRequest and date.

        :param request: OGC-type request with specified bounding box, cloud coverage for specific product.
        :param date: acquisition date or None
        :param size_x: horizontal image dimension
        :param size_y: vertical image dimension
        :return: url to Sentinel Hub's OGC service for this product.
        """
        url = self.get_base_url(request)
        authority = request.theme if hasattr(request, "theme") else self.config.instance_id  # type: ignore

        params = self._get_common_url_parameters(request)
        if request.service_type in (ServiceType.WMS, ServiceType.WCS):
            params = {**params, **self._get_wms_wcs_url_parameters(request, date)}
        if request.service_type is ServiceType.WMS:
            params = {**params, **self._get_wms_url_parameters(request, size_x, size_y)}  # type: ignore
        elif request.service_type is ServiceType.WCS:
            params = {**params, **self._get_wcs_url_parameters(request, size_x, size_y)}  # type: ignore
        elif request.service_type is ServiceType.FIS:
            params = {**params, **self._get_fis_parameters(request, geometry)}  # type: ignore

        return f"{url}/{authority}?{urlencode(params)}"

    def get_base_url(self, request: OgcRequest) -> str:
        """Creates base url string.

        :param request: OGC-type request with specified bounding box, cloud coverage for specific product.
        :return: base string for url to Sentinel Hub's OGC service for this product.
        """
        url = f"{self._base_url}/{request.service_type.value}"  # type: ignore

        if hasattr(request, "data_collection") and request.data_collection.service_url:
            url = url.replace(self.config.sh_base_url, request.data_collection.service_url)

        return url

    @staticmethod
    def _get_common_url_parameters(request: OgcRequest) -> Dict[str, Any]:
        """Returns parameters common dictionary for WMS, WCS and FIS request.

        :param request: OGC-type request with specified bounding box, cloud coverage for specific product.
        :return: dictionary with parameters
        """
        params = {"SERVICE": request.service_type.value, "WARNINGS": False}  # type: ignore

        if hasattr(request, "maxcc"):
            params["MAXCC"] = 100.0 * request.maxcc

        if hasattr(request, "custom_url_params") and request.custom_url_params is not None:
            custom_params = request.custom_url_params

            if CustomUrlParam.EVALSCRIPT in custom_params:
                evalscript = custom_params[CustomUrlParam.EVALSCRIPT]
                custom_params[CustomUrlParam.EVALSCRIPT] = b64encode(evalscript.encode()).decode()

            if CustomUrlParam.GEOMETRY in custom_params:
                geometry = custom_params[CustomUrlParam.GEOMETRY]
                crs = request.bbox.crs

                if isinstance(geometry, Geometry):
                    if geometry.crs is not crs:
                        raise ValueError("Geometry object in custom_url_params should have the same CRS as given BBox")
                else:
                    geometry = Geometry(geometry, crs)

                if geometry.crs is CRS.WGS84:
                    geometry = geometry.reverse()

                custom_params[CustomUrlParam.GEOMETRY] = geometry.wkt

            for resampling in (CustomUrlParam.DOWNSAMPLING, CustomUrlParam.UPSAMPLING):
                if resampling in custom_params:
                    custom_params[resampling] = ResamplingType(custom_params[resampling]).value

            params.update({k.value: str(v) for k, v in custom_params.items()})

        return params

    @staticmethod
    def _get_wms_wcs_url_parameters(request: OgcRequest, date: Optional[datetime.datetime]) -> Dict[str, Any]:
        """Returns parameters common dictionary for WMS and WCS request.

        :param request: OGC-type request with specified bounding box, cloud coverage for specific product.
        :param date: acquisition date or None
        :return: dictionary with parameters
        """
        params = {
            "BBOX": str(request.bbox.reverse()) if request.bbox.crs is CRS.WGS84 else str(request.bbox),
            "FORMAT": MimeType.get_string(request.image_format),
            "CRS": CRS.ogc_string(request.bbox.crs),
        }

        if date is not None:
            start_date = (
                date if request.time_difference < datetime.timedelta(seconds=0) else date - request.time_difference
            )
            end_date = (
                date if request.time_difference < datetime.timedelta(seconds=0) else date + request.time_difference
            )

            start_date_str, end_date_str = serialize_time((start_date, end_date), use_tz=True)
            params["TIME"] = f"{start_date_str}/{end_date_str}"

        return params

    @staticmethod
    def _get_wms_url_parameters(
        request: OgcRequest, size_x: Union[int, str], size_y: Union[int, str]
    ) -> Dict[str, Any]:
        """Returns parameters dictionary for WMS request.

        :param request: OGC-type request with specified bounding box, cloud coverage for specific product.
        :param size_x: horizontal image dimension
        :param size_y: vertical image dimension
        :return: dictionary with parameters
        """
        return {"WIDTH": size_x, "HEIGHT": size_y, "LAYERS": request.layer, "REQUEST": "GetMap", "VERSION": "1.3.0"}

    @staticmethod
    def _get_wcs_url_parameters(
        request: OgcRequest, size_x: Union[int, str], size_y: Union[int, str]
    ) -> Dict[str, Any]:
        """Returns parameters dictionary for WCS request.

        :param request: OGC-type request with specified bounding box, cloud coverage for specific product.
        :param size_x: horizontal image dimension
        :param size_y: vertical image dimension
        :return: dictionary with parameters
        """
        return {"RESX": size_x, "RESY": size_y, "COVERAGE": request.layer, "REQUEST": "GetCoverage", "VERSION": "1.1.2"}

    @staticmethod
    def _get_fis_parameters(request: OgcRequest, geometry: Union[BBox, Geometry]) -> Dict[str, Any]:
        """Returns parameters dictionary for FIS request.

        :param request: OGC-type request with specified bounding box, cloud coverage for specific product.
        :param geometry: list of bounding boxes or geometries
        :return: dictionary with parameters
        """
        start_time, end_time = serialize_time(parse_time_interval(request.time), use_tz=True)

        params = {
            "CRS": CRS.ogc_string(geometry.crs),
            "LAYER": request.layer,
            "RESOLUTION": request.resolution,  # type: ignore
            "TIME": f"{start_time}/{end_time}",
        }

        if not isinstance(geometry, (BBox, Geometry)):
            raise ValueError(
                f"Each geometry must be an instance of sentinelhub.{BBox.__name__} or "
                f"sentinelhub.{Geometry.__name__} but {geometry} found"
            )
        if geometry.crs is CRS.WGS84:
            geometry = geometry.reverse()
        if isinstance(geometry, Geometry):
            params["GEOMETRY"] = geometry.wkt
        else:
            params["BBOX"] = str(geometry)

        if request.bins:  # type: ignore
            params["BINS"] = request.bins  # type: ignore

        if request.histogram_type:  # type: ignore
            params["TYPE"] = request.histogram_type.value  # type: ignore

        return params

    def get_dates(self, request: OgcRequest) -> List[Optional[datetime.datetime]]:
        """Get available Sentinel-2 acquisitions at least time_difference apart

        List of all available Sentinel-2 acquisitions for given bbox with max cloud coverage and the specified
        time interval. When a single time is specified the request will return that specific date, if it exists.
        If a time range is specified the result is a list of all scenes between the specified dates conforming to
        the cloud coverage criteria. Most recent acquisition being first in the list.

        When a time_difference threshold is set to a positive value, the function filters out all datetimes which
        are within the time difference. The oldest datetime is preserved, all others all deleted.

        :param request: OGC-type request
        :return: List of dates of existing acquisitions for the given request
        """
        if request.data_collection.is_timeless:
            return [None]

        if request.wfs_iterator is None:
            self.wfs_iterator = WebFeatureService(
                request.bbox,
                request.time,
                data_collection=request.data_collection,
                maxcc=request.maxcc,
                config=self.config,
            )
        else:
            self.wfs_iterator = request.wfs_iterator

        dates = self.wfs_iterator.get_dates()
        dates = filter_times(dates, request.time_difference)  # type: ignore

        LOGGER.debug("Initializing requests for dates: %s", dates)
        return dates  # type: ignore

    @staticmethod
    def get_image_dimensions(request: OgcRequest) -> Tuple[Union[int, str], Union[int, str]]:
        """Verifies or calculates image dimensions.

        :param request: OGC-type request
        :return: horizontal and vertical dimensions of requested image
        """
        if request.service_type is ServiceType.WCS or (
            isinstance(request.size_x, int) and isinstance(request.size_y, int)
        ):
            return request.size_x, request.size_y  # type: ignore
        if not isinstance(request.size_x, int) and not isinstance(request.size_y, int):
            raise ValueError("At least one of parameters 'width' and 'height' must have an integer value")
        missing_dimension = get_image_dimension(
            request.bbox, width=request.size_x, height=request.size_y  # type: ignore
        )
        if request.size_x is None:
            return missing_dimension, request.size_y  # type: ignore
        if request.size_y is None:
            return request.size_x, missing_dimension
        raise ValueError("Parameters 'width' and 'height' must be integers or None")

    def get_wfs_iterator(self) -> Optional[WebFeatureService]:
        """Returns iterator over info about all satellite tiles used for the request

        :return: Iterator of dictionaries containing info about all satellite tiles used in the request. In case of
            `DataCollection.DEM` it returns `None`.
        """
        return self.wfs_iterator



"""
Module implementing some common utility functions
"""
from enum import Enum
from typing import Any, Dict, Type

from dataclasses_json import LetterCase
from dataclasses_json import config as dataclass_config

from ..geometry import Geometry
from ..time_utils import parse_time, serialize_time

datetime_config = dataclass_config(
    encoder=lambda time: serialize_time(time, use_tz=True) if time else None,
    decoder=lambda time: parse_time(time, force_datetime=True) if time else None,
    letter_case=LetterCase.CAMEL,
)


geometry_config = dataclass_config(
    encoder=Geometry.get_geojson,
    decoder=lambda geojson: Geometry.from_geojson(geojson) if geojson else None,
    exclude=lambda geojson: geojson is None,  # type: ignore[misc, arg-type]
    letter_case=LetterCase.CAMEL,
)


def enum_config(enum_class: Type[Enum]) -> Dict[str, dict]:
    """Given an Enum class it provide an object for serialization/deserialization"""
    return dataclass_config(
        encoder=lambda enum_item: enum_item.value,
        decoder=lambda item: enum_class(item) if item else None,
        exclude=lambda item: item is None,  # type: ignore[misc, arg-type]
        letter_case=LetterCase.CAMEL,
    )


def _update_other_args(dict1: Dict[str, Any], dict2: Dict[str, Any]) -> None:
    """Function for a recursive update of `dict1` with `dict2`. The function loops over the keys in `dict2` and
    only the non-dict like values are assigned to the specified keys.
    """
    for key, value in dict2.items():
        if isinstance(value, dict) and key in dict1:
            _update_other_args(dict1[key], value)
        else:
            dict1[key] = value


def remove_undefined(payload: dict) -> dict:
    """Takes a dictionary and removes keys without value"""
    return {name: value for name, value in payload.items() if value is not None}



"""
Data request interface for downloading satellite data from AWS
"""
from abc import abstractmethod
from typing import Any, Generic, List, Optional, Tuple, TypeVar, Union

from ..base import DataRequest
from ..data_collections import DataCollection
from .client import AwsDownloadClient
from .data import AwsProduct, AwsTile
from .data_safe import SafeProduct, SafeTile

T = TypeVar("T")


class AwsRequest(DataRequest, Generic[T]):
    """The base class for Amazon Web Service request classes. Common parameters are defined here.

    Collects and provides data from AWS.

    More information about Sentinel-2 AWS registry: https://registry.opendata.aws/sentinel-2/
    """

    def __init__(
        self,
        *,
        bands: Union[None, str, List[str]] = None,
        metafiles: Union[None, str, List[str]] = None,
        safe_format: bool = False,
        **kwargs: Any
    ):
        """
        :param bands: List of Sentinel-2 bands for request. If `None` all bands will be obtained
        :param metafiles: list of additional metafiles available on AWS
                          (e.g. ``['metadata', 'tileInfo', 'preview/B01', 'TCI']``)
        :param safe_format: flag that determines the structure of saved data. If `True` it will be saved in .SAFE format
                            defined by ESA. If `False` it will be saved in the same structure as the structure at AWS.
        :param data_folder: location of the directory where the fetched data will be saved.
        :param config: A custom instance of config class to override parameters from the saved configuration.
        """
        self.bands = bands
        self.metafiles = metafiles
        self.safe_format = safe_format

        self.aws_service: T
        super().__init__(AwsDownloadClient, **kwargs)

    @abstractmethod
    def create_request(self) -> None:
        raise NotImplementedError

    def get_aws_service(self) -> T:
        """
        :return: initialized AWS service class
        """
        return self.aws_service


class AwsProductRequest(AwsRequest[AwsProduct]):
    """AWS Service request class for an ESA product."""

    def __init__(self, product_id: str, *, tile_list: Optional[List[str]] = None, **kwargs: Any):
        """
        :param product_id: original ESA product identification string
            (e.g. ``'S2A_MSIL1C_20170414T003551_N0204_R016_T54HVH_20170414T003551'``)
        :param tile_list: list of tiles inside the product to be downloaded. If parameter is set to `None` all
            tiles inside the product will be downloaded.
        :param bands: List of Sentinel-2 bands for request. If `None` all bands will be obtained
        :param metafiles: list of additional metafiles available on AWS
            (e.g. ``['metadata', 'tileInfo', 'preview/B01', 'TCI']``)
        :param safe_format: flag that determines the structure of saved data. If `True` it will be saved in .SAFE format
            defined by ESA. If `False` it will be saved in the same structure as the structure at AWS.
        :param data_folder: location of the directory where the fetched data will be saved.
        :param config: A custom instance of config class to override parameters from the saved configuration.
        """
        self.product_id = product_id
        self.tile_list = tile_list

        super().__init__(**kwargs)

    def create_request(self) -> None:
        if self.safe_format:
            self.aws_service = SafeProduct(
                self.product_id,
                tile_list=self.tile_list,
                bands=self.bands,
                metafiles=self.metafiles,
                config=self.config,
            )
        else:
            self.aws_service = AwsProduct(
                self.product_id,
                tile_list=self.tile_list,
                bands=self.bands,
                metafiles=self.metafiles,
                config=self.config,
            )

        self.download_list, self.folder_list = self.aws_service.get_requests()


class AwsTileRequest(AwsRequest[AwsTile]):
    """AWS Service request class for an ESA tile."""

    def __init__(
        self,
        *,
        data_collection: DataCollection,
        tile: Optional[str] = None,
        time: Optional[str] = None,
        aws_index: Optional[int] = None,
        **kwargs: Any
    ):
        """
        :param data_collection: A collection of requested AWS data. Supported collections are Sentinel-2 L1C and
            Sentinel-2 L2A.
        :param tile: tile name (e.g. ``'T10UEV'``)
        :param time: tile sensing time in ISO8601 format
        :param aws_index: there exist Sentinel-2 tiles with the same tile and time parameter. Therefore, each tile on
            AWS also has an index which is visible in their url path. If aws_index is set to `None` the class
            will try to find the index automatically. If there will be multiple choices it will choose the
            lowest index and inform the user.
        :param bands: List of Sentinel-2 bands for request. If `None` all bands will be obtained
        :param metafiles: list of additional metafiles available on AWS
            (e.g. ``['metadata', 'tileInfo', 'preview/B01', 'TCI']``)
        :param safe_format: flag that determines the structure of saved data. If `True` it will be saved in .SAFE
            format defined by ESA. If `False` it will be saved in the same structure as the structure at AWS.
        :param data_folder: location of the directory where the fetched data will be saved.
        :param config: A custom instance of config class to override parameters from the saved configuration.
        """
        self.data_collection = data_collection
        self.tile = tile
        self.time = time
        self.aws_index = aws_index

        super().__init__(**kwargs)

    def create_request(self) -> None:
        if self.tile is None or self.time is None:
            raise ValueError("The parameters `tile` and `time` must be set.")

        if self.safe_format:
            self.aws_service = SafeTile(
                self.tile,
                self.time,
                self.aws_index,
                bands=self.bands,
                metafiles=self.metafiles,
                data_collection=self.data_collection,
                config=self.config,
            )
        else:
            self.aws_service = AwsTile(
                self.tile,
                self.time,
                self.aws_index,
                bands=self.bands,
                metafiles=self.metafiles,
                data_collection=self.data_collection,
                config=self.config,
            )

        self.download_list, self.folder_list = self.aws_service.get_requests()


def get_safe_format(
    product_id: Optional[str] = None,
    tile: Optional[Tuple[str, str]] = None,
    entire_product: bool = False,
    bands: Union[None, str, List[str]] = None,
    data_collection: Optional[DataCollection] = None,
) -> dict:
    """Returns .SAFE format structure in form of nested dictionaries. Either ``product_id`` or ``tile`` must be
    specified.

    :param product_id: original ESA product identification string. Default is `None`
    :param tile: tuple containing tile name and sensing time/date. Default is `None`
    :param entire_product: in case tile is specified this flag determines if it will be place inside a .SAFE structure
        of the product. Default is `False`
    :param bands: list of bands to download. If `None` all bands will be downloaded. Default is `None`
    :param data_collection: In case of tile request the collection of satellite data has to be specified.
    :return: Nested dictionaries representing .SAFE structure.
    """
    entire_product = entire_product and product_id is None
    if tile is not None:
        safe_tile = SafeTile(tile_name=tile[0], time=tile[1], bands=bands, data_collection=data_collection)
        if not entire_product:
            return safe_tile.get_safe_struct()
        product_id = safe_tile.get_product_id()
    if product_id is None:
        raise ValueError("Either product_id or tile must be specified")
    if entire_product:
        if tile is None:
            raise ValueError("The tile parameter must be set.")
        safe_product = SafeProduct(product_id, tile_list=[tile[0]], bands=bands)
    else:
        safe_product = SafeProduct(product_id, bands=bands)

    return safe_product.get_safe_struct()


def download_safe_format(
    product_id: Optional[str] = None,
    tile: Optional[Tuple[str, str]] = None,
    folder: str = ".",
    redownload: bool = False,
    entire_product: bool = False,
    bands: Optional[List[str]] = None,
    data_collection: Optional[DataCollection] = None,
) -> None:
    """Downloads .SAFE format structure in form of nested dictionaries. Either ``product_id`` or ``tile`` must
    be specified.

    :param product_id: original ESA product identification string. Default is `None`
    :param tile: tuple containing tile name and sensing time/date. Default is `None`
    :param folder: location of the directory where the fetched data will be saved. Default is ``'.'``
    :param redownload: if `True`, download again the requested data even though it's already saved to disk. If
        `False`, do not download if data is already available on disk. Default is `False`
    :param entire_product: in case tile is specified this flag determines if it will be place inside a .SAFE structure
        of the product. Default is `False`
    :param bands: list of bands to download. If `None` all bands will be downloaded. Default is `None`
    :param data_collection: In case of tile request the collection of satellite data has to be specified.
    :return: Nested dictionaries representing .SAFE structure.
    """
    safe_request: Union[None, AwsTileRequest, AwsProductRequest] = None
    entire_product = entire_product and product_id is None
    if tile is not None:
        if data_collection is None:
            raise ValueError("The data_collection parameter must be set.")
        safe_request = AwsTileRequest(
            tile=tile[0],
            time=tile[1],
            data_folder=folder,
            bands=bands,
            safe_format=True,
            data_collection=data_collection,
        )
        if entire_product:
            safe_tile = safe_request.get_aws_service()
            product_id = safe_tile.get_product_id()
    if product_id is not None:
        if entire_product:
            if tile is None:
                raise ValueError("The tile parameter must be set.")
            safe_request = AwsProductRequest(
                product_id, tile_list=[tile[0]], data_folder=folder, bands=bands, safe_format=True
            )
        else:
            safe_request = AwsProductRequest(product_id, data_folder=folder, bands=bands, safe_format=True)

    if safe_request is None:
        raise ValueError("Either 'product_id' or 'tile' has to be defined")

    safe_request.save_data(redownload=redownload)



"""
Implementation of base Sentinel Hub interfaces
"""
from abc import ABCMeta, abstractmethod
from typing import Any, Dict, Optional

from ..base import DataRequest
from ..constants import MimeType, MosaickingOrder, RequestType, ResamplingType
from ..data_collections import DataCollection, OrbitDirection
from ..download import DownloadRequest
from ..geometry import BBox, Geometry
from ..time_utils import RawTimeIntervalType, parse_time_interval, serialize_time
from .utils import _update_other_args


class SentinelHubBaseApiRequest(DataRequest, metaclass=ABCMeta):
    """A base class for Sentinel Hub interfaces"""

    _SERVICE_ENDPOINT = ""
    payload: Dict[str, Any] = {}

    @property
    @abstractmethod
    def mime_type(self) -> MimeType:
        """The mime type of the request."""

    def create_request(self) -> None:
        """Prepares a download request"""
        headers = {"content-type": MimeType.JSON.get_string(), "accept": self.mime_type.get_string()}
        base_url = self._get_base_url()
        self.download_list = [
            DownloadRequest(
                request_type=RequestType.POST,
                url=f"{base_url}/api/v1/{self._SERVICE_ENDPOINT}",
                post_values=self.payload,
                data_folder=self.data_folder,
                save_response=bool(self.data_folder),
                data_type=self.mime_type,
                headers=headers,
                use_session=True,
            )
        ]

    @staticmethod
    def input_data(
        data_collection: DataCollection,
        *,
        identifier: Optional[str] = None,
        time_interval: Optional[RawTimeIntervalType] = None,
        maxcc: Optional[float] = None,
        mosaicking_order: Optional[MosaickingOrder] = None,
        upsampling: Optional[ResamplingType] = None,
        downsampling: Optional[ResamplingType] = None,
        other_args: Optional[Dict[str, Any]] = None,
    ) -> "InputDataDict":
        """Generate the `input data` part of the request body

        :param data_collection: One of supported Process API data collections.
        :param identifier: A collection identifier that can be referred to in the evalscript. Parameter is referenced
            as `"id"` in service documentation. To learn more check
            `data fusion documentation <https://docs.sentinel-hub.com/api/latest/data/data-fusion>`__.
        :param time_interval: A time interval with start and end date of the form YYYY-MM-DDThh:mm:ss or YYYY-MM-DD or
            a datetime object
        :param maxcc: Maximum accepted cloud coverage of an image. Float between 0.0 and 1.0. Default is 1.0.
        :param mosaicking_order: Mosaicking order, which has to be either 'mostRecent', 'leastRecent' or 'leastCC'.
        :param upsampling: A type of upsampling to apply on data
        :param downsampling: A type of downsampling to apply on data
        :param other_args: Additional dictionary of arguments. If provided, the resulting dictionary will get updated
            by it.
        :return: A dictionary-like object that also contains additional attributes
        """
        input_data_dict: Dict[str, Any] = {
            "type": data_collection.api_id,
        }
        if identifier:
            input_data_dict["id"] = identifier

        data_filters = _get_data_filters(data_collection, time_interval, maxcc, mosaicking_order)
        if data_filters:
            input_data_dict["dataFilter"] = data_filters

        processing_params = _get_processing_params(upsampling, downsampling)
        if processing_params:
            input_data_dict["processing"] = processing_params

        if other_args:
            _update_other_args(input_data_dict, other_args)

        return InputDataDict(input_data_dict, service_url=data_collection.service_url)

    @staticmethod
    def bounds(
        bbox: Optional[BBox] = None, geometry: Optional[Geometry] = None, other_args: Optional[Dict[str, Any]] = None
    ) -> Dict[str, Any]:
        """Generate a `bound` part of the API request

        :param bbox: Bounding box describing the area of interest.
        :param geometry: Geometry describing the area of interest.
        :param other_args: Additional dictionary of arguments. If provided, the resulting dictionary will get updated
            by it.
        """
        if bbox is None and geometry is None:
            raise ValueError("'bbox' and/or 'geometry' have to be provided.")

        if bbox and not isinstance(bbox, BBox):
            raise ValueError("'bbox' should be an instance of sentinelhub.BBox")

        if geometry and not isinstance(geometry, Geometry):
            raise ValueError("'geometry' should be an instance of sentinelhub.Geometry")

        if bbox and geometry and bbox.crs != geometry.crs:
            raise ValueError("bbox and geometry should be in the same CRS")

        crs = bbox.crs if bbox else geometry.crs  # type: ignore[union-attr]

        request_bounds: Dict[str, Any] = {"properties": {"crs": crs.opengis_string}}

        if bbox:
            request_bounds["bbox"] = list(bbox)

        if geometry:
            request_bounds["geometry"] = geometry.get_geojson(with_crs=False)

        if other_args:
            _update_other_args(request_bounds, other_args)

        return request_bounds

    def _get_base_url(self) -> str:
        """It decides which base URL to use. Restrictions from data collection definitions overrule the
        settings from config object. In case different collections have different restrictions then
        `SHConfig.sh_base_url` breaks the tie in case it matches one of the data collection URLs.
        """
        data_collection_urls = tuple(
            {
                input_data_dict.service_url.rstrip("/")
                for input_data_dict in self.payload["input"]["data"]
                if isinstance(input_data_dict, InputDataDict) and input_data_dict.service_url is not None
            }
        )
        config_base_url = self.config.sh_base_url.rstrip("/")

        if not data_collection_urls:
            return config_base_url

        if len(data_collection_urls) == 1:
            return data_collection_urls[0]

        if config_base_url in data_collection_urls:
            return config_base_url

        raise ValueError(
            f"Given data collections are restricted to different services: {data_collection_urls}\n"
            "Configuration parameter sh_base_url cannot break the tie because it is set to a different"
            f"service: {config_base_url}"
        )


class InputDataDict(dict):
    """An input data dictionary which also holds additional attributes"""

    def __init__(self, input_data_dict: Dict[str, Any], *, service_url: Optional[str] = None):
        """
        :param input_data_dict: A normal dictionary with input parameters
        :param service_url: A service URL defined by a data collection
        """
        super().__init__(input_data_dict)
        self.service_url = service_url

    def __repr__(self) -> str:
        """Modified dictionary representation that also shows additional attributes"""
        normal_dict_repr = super().__repr__()
        return f"{self.__class__.__name__}({normal_dict_repr}, service_url={self.service_url})"


def _get_data_filters(
    data_collection: DataCollection,
    time_interval: Optional[RawTimeIntervalType],
    maxcc: Optional[float],
    mosaicking_order: Optional[MosaickingOrder],
) -> Dict[str, Any]:
    """Builds a dictionary of data filters for Process API"""
    data_filter: Dict[str, Any] = {}

    if time_interval:
        start_time, end_time = serialize_time(parse_time_interval(time_interval, allow_undefined=True), use_tz=True)
        data_filter["timeRange"] = {"from": start_time, "to": end_time}

    if maxcc is not None:
        if maxcc < 0 or maxcc > 1:
            raise ValueError("maxcc should be a float on an interval [0, 1]")

        data_filter["maxCloudCoverage"] = int(maxcc * 100)

    if mosaicking_order:
        data_filter["mosaickingOrder"] = MosaickingOrder(mosaicking_order).value

    return {**data_filter, **_get_data_collection_filters(data_collection)}


def _get_data_collection_filters(data_collection: DataCollection) -> Dict[str, Any]:
    """Builds a dictionary of filters for Process API from a data collection definition"""
    filters: Dict[str, Any] = {}

    if data_collection.swath_mode:
        filters["acquisitionMode"] = data_collection.swath_mode.upper()

    if data_collection.polarization:
        filters["polarization"] = data_collection.polarization.upper()

    if data_collection.resolution:
        filters["resolution"] = data_collection.resolution.upper()

    if data_collection.orbit_direction and data_collection.orbit_direction.upper() != OrbitDirection.BOTH:
        filters["orbitDirection"] = data_collection.orbit_direction.upper()

    if data_collection.timeliness:
        filters["timeliness"] = data_collection.timeliness

    if data_collection.dem_instance:
        filters["demInstance"] = data_collection.dem_instance

    return filters


def _get_processing_params(
    upsampling: Optional[ResamplingType], downsampling: Optional[ResamplingType]
) -> Dict[str, Any]:
    """Builds a dictionary of processing parameters for Process API"""
    processing_params: Dict[str, Any] = {}

    if upsampling:
        processing_params["upsampling"] = ResamplingType(upsampling).value

    if downsampling:
        processing_params["downsampling"] = ResamplingType(downsampling).value

    return processing_params



"""
The part of the package for interacting with Amazon Web Services (AWS). It requires additional package dependencies
which can be installed with:

.. code-block::

    pip install sentinelhub[AWS]

"""

from .client import AwsDownloadClient
from .constants import AwsConstants
from .data import AwsProduct, AwsTile
from .data_safe import SafeProduct, SafeTile
from .request import AwsProductRequest, AwsTileRequest, download_safe_format, get_safe_format



"""
Module that implements command line interface for AWS package functionalities
"""
from typing import Optional, Tuple

import click

from ..data_collections import DataCollection
from .request import download_safe_format, get_safe_format


@click.command()
@click.option("--product", help="Product ID input")
@click.option("--tile", nargs=2, help="Tile name and date input")
@click.option("-f", "--folder", default=".", help="Set download location")
@click.option("-r", "--redownload", is_flag=True, default=False, help="Redownload existing files")
@click.option("-i", "--info", is_flag=True, default=False, help="Return safe format structure")
@click.option("-e", "--entire", is_flag=True, default=False, help="Get entire product of specified tile")
@click.option("-b", "--bands", default=None, help="Comma separated list (no spaces) of bands to retrieve")
@click.option("--l2a", is_flag=True, default=False, help="In case of tile request this flag specifies L2A products")
def aws(
    product: Optional[str],
    tile: Optional[Tuple[str, str]],
    folder: str,
    redownload: bool,
    info: bool,
    entire: bool,
    bands: Optional[str],
    l2a: bool,
) -> None:
    """Download Sentinel-2 data from Sentinel-2 on AWS to ESA SAFE format. Download uses multiple threads.

    \b
    Examples with Sentinel-2 L1C data:
      sentinelhub.aws --product S2A_MSIL1C_20170414T003551_N0204_R016_T54HVH_20170414T003551
      sentinelhub.aws --product S2A_MSIL1C_20170414T003551_N0204_R016_T54HVH_20170414T003551 -i
      sentinelhub.aws --product S2A_MSIL1C_20170414T003551_N0204_R016_T54HVH_20170414T003551 -f /home/ESA_Products
      sentinelhub.aws --product S2A_MSIL1C_20170414T003551_N0204_R016_T54HVH_20170414T003551 --bands B08,B11
      sentinelhub.aws --tile T54HVH 2017-04-14
      sentinelhub.aws --tile T54HVH 2017-04-14 -e

    \b
    Examples with Sentinel-2 L2A data:
      sentinelhub.aws --product S2A_MSIL2A_20180402T151801_N0207_R068_T33XWJ_20180402T202222
      sentinelhub.aws --tile T33XWJ 2018-04-02 --l2a
    """
    band_list = None if bands is None else bands.split(",")
    data_collection = DataCollection.SENTINEL2_L2A if l2a else DataCollection.SENTINEL2_L1C
    if info:
        if product is None:
            click.echo(get_safe_format(tile=tile, entire_product=entire, data_collection=data_collection))
        else:
            click.echo(get_safe_format(product_id=product))
    else:
        if product is None:
            download_safe_format(
                tile=tile,
                folder=folder,
                redownload=redownload,
                entire_product=entire,
                bands=band_list,
                data_collection=data_collection,
            )
        else:
            download_safe_format(product_id=product, folder=folder, redownload=redownload, bands=band_list)



"""
A download part of the package
"""

from .client import DownloadClient
from .request import DownloadRequest
from .sentinelhub_client import SentinelHubDownloadClient
from .sentinelhub_statistical_client import SentinelHubStatisticalDownloadClient
from .session import SentinelHubSession, SessionSharing, SessionSharingThread, collect_shared_session



"""
Module implementing a rate-limited multithreaded download client for downloading from Sentinel Hub service
"""
import logging
import time
import warnings
from threading import Lock
from typing import Any, Callable, Dict, Optional, Tuple, TypeVar, Union

import requests
from requests import Response

from ..config import SHConfig
from ..exceptions import SHRateLimitWarning, SHRuntimeWarning
from ..type_utils import JsonDict
from .client import DownloadClient
from .handlers import fail_user_errors, retry_temporary_errors
from .rate_limit import SentinelHubRateLimit
from .request import DownloadRequest
from .session import SentinelHubSession

LOGGER = logging.getLogger(__name__)

T = TypeVar("T")


class SentinelHubDownloadClient(DownloadClient):
    """Download client specifically configured for download from Sentinel Hub service"""

    _CACHED_SESSIONS: Dict[Tuple[str, str], SentinelHubSession] = {}
    _UNIVERSAL_CACHE_KEY = "universal-user", "default-url"

    def __init__(self, *, session: Optional[SentinelHubSession] = None, **kwargs: Any):
        """
        :param session: If a session object is provided here then this client instance will always use only the
            provided session. Otherwise, it will either use a cached session or create a new session and cache
            it.
        :param kwargs: Optional parameters from DownloadClient
        """
        super().__init__(**kwargs)

        if session is not None and not isinstance(session, SentinelHubSession):
            raise ValueError(
                f"A session parameter has to be an instance of {SentinelHubSession.__name__} or None, but "
                f"{session} was given"
            )
        self.session = session

        self.rate_limit = SentinelHubRateLimit(num_processes=self.config.number_of_download_processes)
        self.lock: Optional[Lock] = None

    def download(self, *args: Any, **kwargs: Any) -> Any:
        """The main download method

        :param args: Passed to `DownloadClient.download`
        :param kwargs: Passed to `DownloadClient.download`
        """
        # Because the Lock object cannot be pickled we create it only here and remove it afterward
        self.lock = Lock()
        try:
            return super().download(*args, **kwargs)
        finally:
            self.lock = None

    @retry_temporary_errors
    @fail_user_errors
    def _execute_download(self, request: DownloadRequest) -> bytes:
        """
        Executes the download with a single thread and uses a rate limit object, which is shared between all threads
        """
        while True:
            sleep_time = self._execute_thread_safe(self.rate_limit.register_next)

            if sleep_time == 0:
                LOGGER.debug(
                    "Sending %s request to %s. Hash of sent request is %s",
                    request.request_type.value,
                    request.url,
                    request.get_hashed_name(),
                )
                response = self._do_download(request)

                self._execute_thread_safe(self.rate_limit.update, response.headers)

                if response.status_code == requests.status_codes.codes.TOO_MANY_REQUESTS:
                    warnings.warn("Download rate limit hit", category=SHRateLimitWarning)
                    continue

                response.raise_for_status()

                LOGGER.debug("Successful %s request to %s", request.request_type.value, request.url)
                return response.content

            LOGGER.debug("Request needs to wait. Sleeping for %0.2f", sleep_time)
            time.sleep(sleep_time)

    def _execute_thread_safe(self, thread_unsafe_function: Callable[..., T], *args: Any, **kwargs: Any) -> T:
        """Executes a function inside a thread lock and handles potential errors"""
        if self.lock is None:
            return thread_unsafe_function(*args, **kwargs)

        with self.lock:
            return thread_unsafe_function(*args, **kwargs)

    def _do_download(self, request: DownloadRequest) -> Response:
        """Runs the download"""
        if request.url is None:
            raise ValueError(f"Faulty request {request}, no URL specified.")

        return requests.request(
            request.request_type.value,
            url=request.url,
            json=request.post_values,
            headers=self._prepare_headers(request),
            timeout=self.config.download_timeout_seconds,
        )

    def _prepare_headers(self, request: DownloadRequest) -> JsonDict:
        """Prepares final headers by potentially joining them with session headers"""
        if not request.use_session:
            return request.headers

        session_headers = self._execute_thread_safe(self._get_session_headers)
        return {**session_headers, **request.headers}

    def _get_session_headers(self) -> JsonDict:
        """Provides up-to-date session headers

        Note that calling session_headers property triggers update if session has expired therefore this has to be
        called in a thread-safe way
        """
        return self.get_session().session_headers

    def get_session(self) -> SentinelHubSession:
        """Provides the session object used by the client

        :return: A Sentinel Hub session object
        """
        if self.session:
            return self.session

        cache_key = self._get_cache_key(self.config)
        if cache_key in SentinelHubDownloadClient._CACHED_SESSIONS:
            session = SentinelHubDownloadClient._CACHED_SESSIONS[cache_key]
        elif SentinelHubDownloadClient._UNIVERSAL_CACHE_KEY in SentinelHubDownloadClient._CACHED_SESSIONS:
            session = SentinelHubDownloadClient._CACHED_SESSIONS[SentinelHubDownloadClient._UNIVERSAL_CACHE_KEY]
        else:
            session = SentinelHubSession(config=self.config)
            SentinelHubDownloadClient._CACHED_SESSIONS[cache_key] = session

        return session

    @staticmethod
    def cache_session(session: SentinelHubSession, universal: bool = False) -> None:
        """Cache a Sentinel Hub session for to be reused by all instances of `SentinelHubDownloadClient` and its child
        classes within the same Python runtime environment.

        :param session: A session object to be cached.
        :param universal: By default a session is cached for a specific OAuth user ID and Sentinel Hub deployment. But
            if this flag is set to `True` it will cache session for any OAuth user ID and deployment. The intended
            purpose of this parameter is that when a session is sent to a remote processing instance, which doesn't
            have configured Sentinel Hub OAuth credentials, then the session can still be used even without credentials.
        """
        if not isinstance(session, SentinelHubSession):
            raise ValueError(
                f"Given object should be an instance of {SentinelHubSession.__name__} but {session} was given"
            )

        cache_key = (
            SentinelHubDownloadClient._UNIVERSAL_CACHE_KEY
            if universal
            else SentinelHubDownloadClient._get_cache_key(session)
        )
        SentinelHubDownloadClient._CACHED_SESSIONS[cache_key] = session

    @staticmethod
    def _get_cache_key(config_or_session: Union[SentinelHubSession, SHConfig]) -> Tuple[str, str]:
        """Calculates a cache key for the given session or config object. The key consists of an OAuth client ID and
        a base service URL.
        """
        if isinstance(config_or_session, SHConfig):
            return config_or_session.sh_client_id, config_or_session.sh_base_url

        if isinstance(config_or_session, SentinelHubSession):
            base_url = config_or_session.config.sh_base_url

            # If session was generated from token then config_or_session.config.sh_client_id could have wrong client id.
            sh_client_id = config_or_session.info().get("aud", "")
            if not sh_client_id:
                warnings.warn(
                    "Failed to read client ID from OAuth token. Session caching might not work correctly.",
                    category=SHRuntimeWarning,
                )

            return sh_client_id, base_url

        raise ValueError(f"Expected a config or a session object but got {config_or_session}")

    @staticmethod
    def clear_cache() -> None:
        """Clears cached sessions."""
        SentinelHubDownloadClient._CACHED_SESSIONS = {}



"""
Module implementing the main download client class
"""
import json
import logging
import os
import sys
import warnings
from concurrent.futures import Future, ThreadPoolExecutor, as_completed
from typing import Any, List, Optional, Union, cast, overload
from xml.etree import ElementTree

import requests
from tqdm.auto import tqdm

from ..config import SHConfig
from ..constants import MimeType, RequestType
from ..decoding import decode_data as decode_data_function
from ..exceptions import (
    DownloadFailedException,
    HashedNameCollisionException,
    MissingDataInRequestException,
    SHRuntimeWarning,
)
from ..io_utils import read_data, write_data
from ..type_utils import JsonDict
from .handlers import fail_user_errors, retry_temporary_errors
from .request import DownloadRequest

LOGGER = logging.getLogger(__name__)


class DownloadClient:
    """A basic download client object

    It does the following:

    - downloads the data with multiple threads in parallel,
    - handles any exceptions that occur during download,
    - decodes downloaded data,
    - reads and writes locally stored/cached data
    """

    def __init__(
        self, *, redownload: bool = False, raise_download_errors: bool = True, config: Optional[SHConfig] = None
    ):
        """
        :param redownload: If `True` the data will always be downloaded again. By default, this is set to `False` and
            the data that has already been downloaded and saved to an expected location will be read from the
            location instead of being downloaded again.
        :param raise_download_errors: If `True` any error in download process will be raised as
            `DownloadFailedException`. If `False` failed downloads will only raise warnings.
        :param config: An instance of configuration class
        """
        self.redownload = redownload
        self.raise_download_errors = raise_download_errors

        self.config = config or SHConfig()

    @overload
    def download(
        self,
        download_requests: DownloadRequest,
        max_threads: Optional[int] = None,
        decode_data: bool = True,
        show_progress: bool = False,
    ) -> Any:
        ...

    @overload
    def download(
        self,
        download_requests: List[DownloadRequest],
        max_threads: Optional[int] = None,
        decode_data: bool = True,
        show_progress: bool = False,
    ) -> List[Any]:
        ...

    def download(
        self,
        download_requests: Union[DownloadRequest, List[DownloadRequest]],
        max_threads: Optional[int] = None,
        decode_data: bool = True,
        show_progress: bool = False,
    ) -> Union[List[Any], Any]:
        """Download one or multiple requests, provided as a request list.

        :param download_requests: A list of requests or a single request to be executed.
        :param max_threads: Maximum number of threads to be used for download in parallel. The default is
            `max_threads=None` which will use the number of processors on the system multiplied by 5.
        :param decode_data: If `True` it will decode data otherwise it will return it in binary format.
        :param show_progress: Whether a progress bar should be displayed while downloading
        :return: A list of results or a single result, depending on input parameter `download_requests`
        """
        downloads = [download_requests] if isinstance(download_requests, DownloadRequest) else download_requests

        data_list = [None] * len(downloads)

        with ThreadPoolExecutor(max_workers=max_threads) as executor:
            download_list = [executor.submit(self._single_download, request, decode_data) for request in downloads]
            future_order = {future: i for i, future in enumerate(download_list)}

            # Consider using tqdm.contrib.concurrent.thread_map in the future
            if show_progress:
                with tqdm(total=len(download_list)) as pbar:
                    for future in as_completed(download_list):
                        data_list[future_order[future]] = self._process_download_future(future)
                        pbar.update(1)
            else:
                for future in as_completed(download_list):
                    data_list[future_order[future]] = self._process_download_future(future)

        if isinstance(download_requests, DownloadRequest):
            return data_list[0]
        return data_list

    def _process_download_future(self, future: Future) -> Any:
        """Unpacks the future and correctly handles exceptions"""
        try:
            return future.result()
        except DownloadFailedException as download_exception:
            if self.raise_download_errors:
                traceback = sys.exc_info()[2]
                raise download_exception.with_traceback(traceback)

            warnings.warn(str(download_exception), category=SHRuntimeWarning)
            return None

    def _single_download(self, request: DownloadRequest, decode_data: bool) -> Any:
        """Method for downloading a single request"""
        request.raise_if_invalid()

        request_path, response_path = request.get_storage_paths()

        if not self._is_download_required(request, response_path):
            path = cast(str, response_path)  # refactor in future, for now ensured by _is_download_required
            if request.return_data:
                LOGGER.debug("Reading locally stored data from %s instead of downloading", path)
                self._check_cached_request_is_matching(request, request_path)
                return read_data(path, data_format=request.data_type if decode_data else MimeType.RAW)
            return None

        response_content = self._execute_download(request)

        if request_path and request.save_response and (self.redownload or not os.path.exists(request_path)):
            request_info = request.get_request_params(include_metadata=True)
            write_data(request_path, request_info, MimeType.JSON)
            LOGGER.debug("Saved request info to %s", request_path)

        if request.save_response:
            path = cast(str, response_path)  # refactor in future, for now ensured by raise_if_invalid
            write_data(path, response_content, MimeType.RAW)
            LOGGER.debug("Saved data to %s", path)

        if request.return_data:
            if decode_data:
                return decode_data_function(response_content, request.data_type)
            return response_content
        return None

    @retry_temporary_errors
    @fail_user_errors
    def _execute_download(self, request: DownloadRequest) -> bytes:
        """A default way of executing a single download request"""
        if request.url is None:
            raise ValueError(f"Faulty request {request}, no URL specified.")

        LOGGER.debug(
            "Sending %s request to %s. Hash of sent request is %s",
            request.request_type.value,
            request.url,
            request.get_hashed_name(),
        )

        response = requests.request(
            request.request_type.value,
            url=request.url,
            json=request.post_values,
            headers=request.headers,
            timeout=self.config.download_timeout_seconds,
        )

        response.raise_for_status()
        LOGGER.debug("Successful %s request to %s", request.request_type.value, request.url)

        return response.content

    def _is_download_required(self, request: DownloadRequest, response_path: Optional[str]) -> bool:
        """Checks if download should actually be done"""
        return (request.save_response or request.return_data) and (
            self.redownload or response_path is None or not os.path.exists(response_path)
        )

    @staticmethod
    def _check_cached_request_is_matching(request: DownloadRequest, request_path: Optional[str]) -> None:
        """Ensures that the cached request matches the current one. Serves as protection against hash collisions"""
        if not request_path:
            return

        cached_request_info = read_data(request_path, MimeType.JSON)
        current_request_info = request.get_request_params(include_metadata=False)
        # Timestamps are allowed to differ
        del cached_request_info["timestamp"]
        del cached_request_info["headers"]
        # Saved request was jsonified
        current_request_info_json = json.loads(json.dumps(current_request_info))

        if cached_request_info != current_request_info_json:
            raise HashedNameCollisionException(
                f"Request has hashed name {request.get_hashed_name()}, which matches request saved at {request_path}, "
                "but the requests are different. Possible hash collision"
            )

    def get_json(
        self,
        url: str,
        post_values: Optional[JsonDict] = None,
        headers: Optional[JsonDict] = None,
        request_type: Optional[RequestType] = None,
        **kwargs: Any,
    ) -> Union[JsonDict, list, str, None]:
        """Download request as JSON data type

        :param url: A URL from where the data will be downloaded
        :param post_values: A dictionary of parameters for a POST request
        :param headers: A dictionary of additional request headers
        :param request_type: A type of HTTP request to make. If not specified, then it will be a GET request if
            `post_values=None` and a POST request otherwise
        :param kwargs: Any other parameters that are passed to DownloadRequest class
        :return: JSON data parsed into Python objects
        """
        json_headers = headers or {}

        if request_type is None:
            request_type = RequestType.GET if post_values is None else RequestType.POST

        if request_type is RequestType.POST and "Content-Type" not in json_headers:
            json_headers = {"Content-Type": MimeType.JSON.get_string(), **json_headers}

        request = DownloadRequest(
            url=url,
            headers=json_headers,
            request_type=request_type,
            post_values=post_values,
            data_type=MimeType.JSON,
            **kwargs,
        )

        return self._single_download(request, decode_data=True)

    def get_json_dict(self, url: str, *args: Any, extract_key: Optional[str] = None, **kwargs: Any) -> JsonDict:
        """Download request as JSON data type, failing if the result is not a dictionary

        For other parameters see `get_json` method.

        :param url: A URL from where the data will be downloaded
        :param extract_key: If provided, the field is automatically extracted, checked, and returned
        """
        response = self.get_json(url, *args, **kwargs)

        if not isinstance(response, dict):
            raise MissingDataInRequestException(
                f"Response from {url} was expected to be a dictionary, but got {type(response)}."
            )

        if extract_key is None:
            return response
        if extract_key in response and isinstance(response[extract_key], dict):
            return response[extract_key]

        explanation = f"Value is of type {type(response[extract_key])}" if extract_key in response else "Key is missing"
        raise MissingDataInRequestException(
            f"Response from {url} was expected to have {extract_key} mapping to a dictionary. {explanation}."
        )

    def get_xml(self, url: str, **kwargs: Any) -> ElementTree.ElementTree:
        """Download request as XML data type

        :param url: url to Sentinel Hub's services or other sources from where the data is downloaded
        :param kwargs: Any other parameters that are passed to DownloadRequest class
        :return: request response as XML instance
        """
        request = DownloadRequest(url=url, data_type=MimeType.XML, **kwargs)
        return self._single_download(request, decode_data=True)



"""
A client interface for `Sentinel Hub Catalog API <https://docs.sentinel-hub.com/api/latest/api/catalog>`__.
"""
import datetime as dt
from typing import Any, Iterable, List, Optional, Union

from ..base import FeatureIterator
from ..data_collections import DataCollection, OrbitDirection
from ..geometry import CRS, BBox, Geometry
from ..time_utils import parse_time, parse_time_interval, serialize_time
from ..type_utils import JsonDict, RawTimeIntervalType, RawTimeType
from .base import SentinelHubService
from .utils import remove_undefined


class SentinelHubCatalog(SentinelHubService):
    """The main class for interacting with Sentinel Hub Catalog API

    For more details about certain endpoints and parameters check
    `Catalog API documentation <https://docs.sentinel-hub.com/api/latest/api/catalog>`__.
    """

    @staticmethod
    def _get_service_url(base_url: str) -> str:
        """Provides URL to Catalog API"""
        return f"{base_url}/api/v1/catalog"

    def get_info(self) -> JsonDict:
        """Provides the main information that define Sentinel Hub Catalog API

        `Catalog API reference <https://docs.sentinel-hub.com/api/latest/reference/#operation/getLandingPage>`__

        :return: A service payload with information
        """
        return self.client.get_json_dict(self.service_url)

    def get_conformance(self) -> JsonDict:
        """Get information about specifications that this API conforms to

        `Catalog API reference
        <https://docs.sentinel-hub.com/api/latest/reference/#operation/getConformanceDeclaration>`__

        :return: A service payload with information
        """
        return self.client.get_json_dict(f"{self.service_url}/conformance")

    def get_collections(self) -> List[JsonDict]:
        """Provides a list of collections that are available to a user

        `Catalog API reference <https://docs.sentinel-hub.com/api/latest/reference/#operation/getCollections>`__

        :return: A list of collections with information
        """
        return self.client.get_json_dict(f"{self.service_url}/collections", use_session=True)["collections"]

    def get_collection(self, collection: Union[DataCollection, str]) -> JsonDict:
        """Provides information about given collection

        `Catalog API reference <https://docs.sentinel-hub.com/api/latest/reference/#operation/describeCollection>`__

        :param collection: A data collection object or a collection ID
        :return: Information about a collection
        """
        collection_id = self._parse_collection_id(collection)
        url = f"{self.service_url}/collections/{collection_id}"
        return self.client.get_json_dict(url, use_session=True)

    def get_feature(self, collection: DataCollection, feature_id: str) -> JsonDict:
        """Provides information about a single feature in a collection

        `Catalog API reference <https://docs.sentinel-hub.com/api/latest/reference/#operation/getFeature>`__

        :param collection: A data collection object or a collection ID
        :param feature_id: A feature ID
        :return: Information about a feature in a collection
        """
        collection_id = self._parse_collection_id(collection)
        url = f"{self.service_url}/collections/{collection_id}/items/{feature_id}"
        return self.client.get_json_dict(url, use_session=True)

    def search(
        self,
        collection: Union[DataCollection, str],
        *,
        time: Union[RawTimeType, RawTimeIntervalType] = None,
        bbox: Optional[BBox] = None,
        geometry: Optional[Geometry] = None,
        ids: Optional[List[str]] = None,
        query: Optional[JsonDict] = None,
        fields: Optional[JsonDict] = None,
        distinct: Optional[str] = None,
        limit: int = 100,
        **kwargs: Any,
    ) -> "CatalogSearchIterator":
        """Catalog STAC search

        `Catalog API reference <https://docs.sentinel-hub.com/api/latest/reference/#operation/postSearchSTAC>`__

        :param collection: A data collection object or a collection ID
        :param time: A time interval or a single time. It can either be a string in form  YYYY-MM-DDThh:mm:ss or
            YYYY-MM-DD or a datetime object
        :param bbox: A search bounding box, it will always be reprojected to WGS 84 before being sent to the service.
            Re-projection will be done with BBox.transform_bounds method which can produce a slightly larger bounding
            box. If that is a problem then transform a BBox object into a Geometry object and search with geometry
            parameter instead.
        :param geometry: A search geometry, it will always reprojected to WGS 84 before being sent to the service.
            This parameter is defined with parameter `intersects` at the service.
        :param ids: A list of feature ids as defined in service documentation
        :param query: A STAC query described in Catalog API documentation
        :param fields: A dictionary of fields to include or exclude described in Catalog API documentation
        :param distinct: A special query attribute described in Catalog API documentation
        :param limit: A number of results to return per each request. At the end iterator will always provide all
            results the difference is only in how many requests it will have to make in the background.
        :param kwargs: Any other parameters that will be passed directly to the service
        """
        url = f"{self.service_url}/search"

        collection_id = self._parse_collection_id(collection)
        start_time, end_time = serialize_time(parse_time_interval(time, allow_undefined=True), use_tz=True)

        if bbox and bbox.crs is not CRS.WGS84:
            bbox = bbox.transform_bounds(CRS.WGS84)
        if geometry and geometry.crs is not CRS.WGS84:
            geometry = geometry.transform(CRS.WGS84)

        _query = self._get_data_collection_filters(collection)
        if query:
            _query.update(query)

        payload = remove_undefined(
            {
                "collections": [collection_id],
                "datetime": f"{start_time}/{end_time}" if time else None,
                "bbox": list(bbox) if bbox else None,
                "intersects": geometry.get_geojson(with_crs=False) if geometry else None,
                "ids": ids,
                "query": _query,
                "fields": fields,
                "distinct": distinct,
                "limit": limit,
                **kwargs,
            }
        )

        return CatalogSearchIterator(self.client, url, payload)

    @staticmethod
    def _parse_collection_id(collection: Union[str, DataCollection]) -> str:
        """Extracts catalog collection id from an object defining a collection"""
        if isinstance(collection, DataCollection):
            return collection.catalog_id
        if isinstance(collection, str):
            return collection
        raise ValueError(f"Expected either a DataCollection object or a collection id string, got {collection}")

    @staticmethod
    def _get_data_collection_filters(data_collection: Union[DataCollection, str]) -> JsonDict:
        """Builds a dictionary of query filters for catalog API from a data collection definition"""
        filters: JsonDict = {}

        if isinstance(data_collection, str):
            return filters

        if data_collection.swath_mode:
            filters["sar:instrument_mode"] = {"eq": data_collection.swath_mode.upper()}

        if data_collection.polarization:
            filters["polarization"] = {"eq": data_collection.polarization.upper()}

        if data_collection.resolution:
            filters["resolution"] = {"eq": data_collection.resolution.upper()}

        if data_collection.orbit_direction and data_collection.orbit_direction.upper() != OrbitDirection.BOTH:
            filters["sat:orbit_state"] = {"eq": data_collection.orbit_direction.lower()}

        if data_collection.timeliness:
            filters["timeliness"] = {"eq": data_collection.timeliness}

        return filters


class CatalogSearchIterator(FeatureIterator[JsonDict]):
    """Searches a catalog with a given query and provides results"""

    def __init__(self, *args: Any, **kwargs: Any):
        super().__init__(*args, **kwargs)
        self.next: Optional[JsonDict] = None

    def _fetch_features(self) -> Iterable[JsonDict]:
        """Collects more results from the service"""
        payload = remove_undefined({**self.params, "next": self.next})

        results = self.client.get_json_dict(self.url, post_values=payload, use_session=True)

        self.next = results["context"].get("next")
        new_features = results["features"]
        self.finished = self.next is None or not new_features

        return new_features

    def get_timestamps(self) -> List[dt.date]:
        """Provides features timestamps

        :return: A list of sensing times
        """
        return [parse_time(feature["properties"]["datetime"]) for feature in self]  # type: ignore[misc]

    def get_geometries(self) -> List[Geometry]:
        """Provides features geometries

        :return: A list of geometry objects with CRS
        """
        return [Geometry.from_geojson(feature["geometry"]) for feature in self]

    def get_ids(self) -> List[str]:
        """Provides features IDs

        :return: A list of IDs
        """
        return [feature["id"] for feature in self]



"""
Module for working with Sentinel Hub FIS service
"""
import datetime
import warnings
from typing import Any, List, Optional, Union

from ..constants import HistogramType, MimeType, RequestType, ServiceType, SHConstants
from ..download import DownloadRequest
from ..exceptions import SHDeprecationWarning
from ..geometry import BBox, Geometry
from ..time_utils import RawTimeIntervalType, RawTimeType
from .ogc import OgcImageService, OgcRequest
from .wfs import WebFeatureService


class FisRequest(OgcRequest):
    """``Deprecated - use Statistical API instead!``

    The class for interacting with Feature Info Service.

    For more info check `FIS documentation <https://www.sentinel-hub.com/develop/api/ogc/fis-request/>`__.
    """

    def __init__(
        self,
        layer: str,
        time: Union[RawTimeType, RawTimeIntervalType],
        geometry_list: List[Union[Geometry, BBox]],
        *,
        resolution: str = "10m",
        bins: Optional[str] = None,
        histogram_type: Optional[HistogramType] = None,
        **kwargs: Any,
    ):
        """
        :param layer: An ID of a layer configured in Sentinel Hub Dashboard. It has to be configured for the same
            instance ID which will be used for this request. The satellite collection of the layer in Dashboard
            must also match the one given by `data_collection` parameter.
        :param time: time or time range for which to return the results, in ISO8601 format
            (year-month-date, for example: ``2016-01-01``, or year-month-dateThours:minutes:seconds format,
            i.e. ``2016-01-01T16:31:21``). Examples: ``'2016-01-01'``, or ``('2016-01-01', ' 2016-01-31')``
            (datetime.datetime, datetime.datetime)
        :param geometry_list: A WKT representation of a geometry describing the region of interest.
            Note that WCS 1.1.1 standard is used here, so for EPSG:4326 coordinates should be in latitude/longitude
            order.
        :param resolution: Specifies the spatial resolution, in meters per pixel, of the image from which the statistics
            are to be estimated. When using CRS=EPSG:4326 one has to add the "m" suffix to
            enforce resolution in meters per pixel (e.g. RESOLUTION=10m).
        :param bins: The number of bins (a positive integer) in the histogram. If this parameter is absent no histogram
            is computed.
        :param histogram_type: type of histogram
        :param data_collection: A collection of requested satellite data. It has to be the same as defined in Sentinel
            Hub Dashboard for the given layer. Default is Sentinel-2 L1C.
        :param maxcc: maximum accepted cloud coverage of an image. Float between 0.0 and 1.0. Default is ``1.0``.
        :param custom_url_params: Dictionary of CustomUrlParameters and their values supported by Sentinel Hub's WMS
            and WCS services. All available parameters are described in
            `documentation <https://www.sentinel-hub.com/develop/api/ogc/custom-parameters/>`__. Note: in
            case of constants.CustomUrlParam.EVALSCRIPT the dictionary value must be a string
            of Javascript code that is not encoded into base64.
        :param data_folder: location of the directory where the fetched data will be saved.
        :param config: A custom instance of config class to override parameters from the saved configuration.
        """
        self.geometry_list = geometry_list
        self.resolution = resolution
        self.bins = bins
        self.histogram_type = HistogramType(histogram_type) if histogram_type else None

        super().__init__(bbox=None, layer=layer, time=time, service_type=ServiceType.FIS, **kwargs)  # type: ignore

    def create_request(self) -> None:  # type: ignore
        """Set download requests

        Create a list of DownloadRequests for all Sentinel-2 acquisitions within request's time interval and
        acceptable cloud coverage.
        """
        fis_service = _FisService(config=self.config)
        self.download_list = fis_service.get_request(self)

    def get_dates(self) -> List[Optional[datetime.datetime]]:
        """This method is not supported for FIS request"""
        raise NotImplementedError

    def get_tiles(self) -> Optional[WebFeatureService]:
        """This method is not supported for FIS request"""
        raise NotImplementedError


class _FisService(OgcImageService):
    """Sentinel Hub OGC services class for providing FIS data

    Intermediate layer between FIS requests and the Sentinel Hub FIS services.
    """

    def __init__(self, *args: Any, **kwargs: Any):
        message = (
            "Fis service is being deprecated in favour of SentinelHubStatistical. "
            "Although no immediate action is needed as FIS is still supported, consider switching to "
            "Statistical API because it provides additional functionalities."
        )
        warnings.warn(message, category=SHDeprecationWarning)

        super().__init__(*args, **kwargs)

    def get_request(self, request: FisRequest) -> List[DownloadRequest]:  # type: ignore
        """Get download requests

        Create a list of DownloadRequests for all Sentinel-2 acquisitions within request's time interval and
        acceptable cloud coverage.

        :param request: OGC-type request with specified bounding box, time interval, and cloud coverage for specific
            product.
        """

        return [self._create_request(request=request, geometry=geometry) for geometry in request.geometry_list]

    def _create_request(self, request: FisRequest, geometry: Union[BBox, Geometry]) -> DownloadRequest:
        url = self.get_base_url(request)

        headers = {"Content-Type": MimeType.JSON.get_string(), **SHConstants.HEADERS}

        post_data = {**self._get_common_url_parameters(request), **self._get_fis_parameters(request, geometry)}
        post_data = {k.lower(): v for k, v in post_data.items()}  # lowercase required on SH service

        return DownloadRequest(
            url=f"{url}/{self.config.instance_id}",
            post_values=post_data,
            data_type=MimeType.JSON,
            headers=headers,
            request_type=RequestType.POST,
        )



"""
Utility tools for writing unit tests for packages which rely on `sentinelhub-py`
"""
import os
from typing import Optional, Tuple

import numpy as np
from pytest import approx


def get_input_folder(current_file: str) -> str:
    """Use fixtures if possible. This is meant only for test cases"""
    return os.path.join(os.path.dirname(os.path.realpath(current_file)), "TestInputs")


def get_output_folder(current_file: str) -> str:
    """Use fixtures if possible. This is meant only for test cases"""
    return os.path.join(os.path.dirname(os.path.realpath(current_file)), "TestOutputs")


def test_numpy_data(
    data: Optional[np.ndarray] = None,
    exp_shape: Optional[Tuple[int, ...]] = None,
    exp_dtype: Optional[np.dtype] = None,
    exp_min: Optional[float] = None,
    exp_max: Optional[float] = None,
    exp_mean: Optional[float] = None,
    exp_median: Optional[float] = None,
    exp_std: Optional[float] = None,
    delta: Optional[float] = None,
) -> None:
    """Validates basic statistics of data array

    :param data: Data array
    :param exp_shape: Expected shape
    :param exp_dtype: Expected dtype
    :param exp_min: Expected minimal value
    :param exp_max: Expected maximal value
    :param exp_mean: Expected mean value
    :param exp_median: Expected median value
    :param exp_std: Expected standard deviation value
    :param delta: Precision of validation (relative). If not set, it will be set automatically
    """
    if data is None:
        return
    delta = delta if delta is not None else 1e-4

    stats_suite = {
        "shape": (lambda array: array.shape, exp_shape),
        "dtype": (lambda array: array.dtype, exp_dtype),
        "min": (np.nanmin, exp_min),
        "max": (np.nanmax, exp_max),
        "mean": (np.nanmean, exp_mean),
        "median": (np.nanmedian, exp_median),
        "std": (np.nanstd, exp_std),
    }

    is_precise = {"shape", "dtype"}

    data_stats, exp_stats = {}, {}
    for name, (func, expected) in stats_suite.items():
        if expected is not None:
            data_stats[name] = func(data)  # type: ignore # unknown function
            exp_stats[name] = expected if name in is_precise else approx(expected, rel=delta)

    assert data_stats == exp_stats, "Statistics differ from expected values"



"""
The part of the package for interacting with Geopedia API. Note that this part might in the future be moved into a
separate package!
"""

from .core import GeopediaFeatureIterator, GeopediaSession
from .request import GeopediaImageRequest, GeopediaWmsRequest
